
/* Helper program for running GENECONV */

/* GENECONV_HELPER communicates with GENECONV with a */
/*   `system(...)' command. This function does not appear to be */
/*   ANSI standard, but is supported by most C/C++ compilers. */

/* `version.c' is linked in, which contains the name of various */
/*   program (GENECONV and GENECONV_HELPER), the program version, etc */

#include "vall.h"
#include <errno.h>

#define COMMCHARS "[#;"
#define CONFIG_MAGIC "#GCONV_CONFIG"

#define OPC  5
#define SPC  (OPC+5)       /* For menu spacing */

#define LBUFF 80           /* Edit window buffer sizes */
#define LSIZE (LBUFF+50)

/* UNIX/MSDOS constants (MSDOS is Win 95/98/NT/XP etc) */
#ifdef WIN32_DOS
  #define DIRSEPCHAR ';'
  #define DIRSEPSTRING ";"
  #define PATHSEPCHAR '\\'
#else /* UNIX and MACs */
  #define DIRSEPCHAR ':'
  #define DIRSEPSTRING ":"
  #define PATHSEPCHAR '/'
#endif /* WIN32_DOS */

static char *ini_name, *prog_name, *newpath, *inpath;
static int linenum, newdefs;
static char *mainprog_execname;

/* Buffer strings from INI file */
static char *seqnbuff, *filepath, *outpath, *optbuff;

/* Output file names and paths */
static char *seqname, *outname;

/* Option lists */
typedef struct { char **list;
                 int n,nx, nf,nc; } STRLIST, *PSTRLIST;

static STRLIST sql, opt, vars;

/* Structures and data for INI file */

typedef enum { INOTYPE, ISEQNAME, IPATH, ILIST }  INITYPE;

typedef struct { char *tag, **vname;
                 INITYPE initype;  STRLIST *sp;
                 char *verb; }  INIFLAG;

static INIFLAG iniflag[] = {
   { "Seqname", &seqnbuff, ILIST, &sql },
   { "Inputdatapath", &filepath, IPATH, NULL, "-filepath=" },
   { "Outpath", &outpath, IPATH, NULL, "-outpath=" },
   { "Options", &optbuff, ILIST, &opt },
   { NULL } };


/* Prototypes of principal functions */

static void Enter_seqnames(void);
static void Enter_opts(void);
static void show_current_menu_filename(const char *fname);
static void Enter_filepath(void);
static void Enter_outpath(void);

static void Opt_help(void);
static void Clear_options(void);
static void Run_Gconv(void);
static void Show_commline(void);
static void Write_config(void);

static void setpath(const char *inpath);
static void read_ini(void);
static STRLIST new_strlist(void);
static char *pack_strlist(STRLIST s);
static char *parse_path(const char *fbuff, const char *bleep);
static void parse_vars(void);
static void say_opts(void);
static void say_opts_out(void);
static char *warn_nofc(void);
static void kill_strlist(PSTRLIST sp);
static int need_quotes(const char *str);
static char *do_quotes(const char *str);
static char *addstr (PSTRLIST sp, const char *str);
static int parse_add(PSTRLIST sp, const char *sbuff, int *nokp);
static void set_names(void);

static void write_ini(void);
static void ex_out(void);
static void free_all(void);

static char *lineread(FILE *rff);

/* Utilities in this program */
/* The GENECONV utility file `vutil.c' cannot be linked in */
/*   because it is compiled under different parameters. */

static FILE *fxopen(const char *path, const char *mode);
static int  fxclose(FILE *fp, const char *fname);
static int searchexe(const char *fname);

static int getline (char *buff, int n);
static void nl(void);
static int get2char (void);
static void presskey (void);

static void *cxalloc (size_t nitems, size_t size);
static void *rexxalloc(void *p, size_t ssize, size_t oldnum, size_t newnum);
static void *mxalloc (size_t size);
static char *strxdup(const char *p);
static char *strxndup(const char *p, size_t n);
static int strmatch(const char *str, const char *mat);
static int stricmp(const char *p, const char *q);
static int strxcmp(const char *str, const char *mat);
static char *makeuc(char *str);

typedef struct { char *drive, *path, *name, *ext; }  PARTFN;

static PARTFN partfn (const char *source);
static char *makefnam (const char *rawfn, const char *tplate);
static char *makefmpf (PARTFN pf);
static char *makefullpath (PARTFN pf);
static void freepf (PARTFN pf);
static char *addstrip(char *str1, int addsep);
static char *addstripnew(const char *str1, int addsep);
static int isextent(const char *p, const char *ext);
static char *add_extent(const char *fname, const char *ext);

static int utoklist (char **vars, int nv, int col, int collim, int rgap);
static int ftoklist (FILE *wf, char **vars, int nv,
   int col, int collim, int rgap);
static int stoklist (STRLIST s, int col);
static int sdtoklist (STRLIST s);

static int ask(const char *prompt, const char* anslist);
static int askd(const char *prompt, const char* anslist);


/* Main program */

int main(int argc, char **argv)
 { PARTFN pf;  int firflag=TRUE;
   printf ("\n %s Helper Program -- Vs. %s\n", MainProgName, ProgVersion);
   pf=partfn(argv[0]);
   prog_name=(pf.name[0])?makeuc(strxdup(pf.name)):HelperName;
   /* Try to make sure that the path includes argv[0] */
   if (pf.path[0]!=0 && argc>=-1)
    { inpath=makefullpath(pf);
      setpath(inpath);
      printf (" %s was found in the directory %s\n",
        prog_name, inpath); }
   freepf(pf);
   /* `mainprogname' should be `geneconv', but without */
   /*   the extension `exe' */
   #ifdef WIN32_DOS
   if (isextent(mainprogname,"exe"))
      mainprog_execname=strxdup(mainprogname);
   else  mainprog_execname=add_extent(mainprogname,".exe");
   #else
   mainprog_execname=strxdup(mainprogname);
   #endif

   /* Read  Geneconv_Helper.ini  for default values of variables */
   /*   seqnbuff, optbuff, filepath, and outpath  */
   /* Parse the first 3 into lists */
   ini_name=makefnam(".ini",argv[0]);
   read_ini();
   if (parse_add(&sql,seqnbuff,NULL))  newdefs=TRUE;
   if (parse_add(&opt,optbuff,NULL))  newdefs=TRUE;

   /* See if these lists contain an obvious sequence file name, */
   /*   and set `outpath' if needed */
   set_names();

   /* Main menu */
   for(;;)
    { int ncurr=0;
      if (firflag) firflag=FALSE;  else nl();
      printf ("\n MAIN MENU:\n");
      printf ("%*sA. Enter Seqfile file name "
        "and/or configuration files.\n", OPC,"");
      if (sql.n>0)
       { int col=printf ("%*sCurrent entry:  ",  SPC,"");
         stoklist(sql,col);  ncurr++; }
      else  printf ("%*sNo current entries\n",  SPC,"");
      if (sql.n>0 && sql.nf==0 && sql.nc==0)
        printf ("%*s%s\n", SPC,"", warn_nofc());
      printf ("%*sB. Enter options\n", OPC,"");
      if (opt.n>0)
       { int col=printf ("%*sCurrent option list:  ", SPC,"");
         stoklist(opt,col);  ncurr++; }
      else  printf ("%*sNo current entries\n",  SPC,"");
      printf ("%*sF. Optional input file path for <Seqfiles> "
           "and config files.\n", OPC,"");
      show_current_menu_filename(filepath);
      if (filepath!=NULL) ncurr++;
      printf ("%*sO. Change output file directory\n", OPC,"");
      show_current_menu_filename((outpath)?outpath:"NONE");
      printf ("%*sH. Hints for additional options\n", OPC,"");
      if (ncurr>0)
        printf ("%*sZ. Clear most current entries\n", OPC,"");
      nl();
      printf ("%*sR. Run %s with current options\n", OPC,"", MainProgName);
      printf ("%*sS. Show the command line so far\n", OPC,"");
      if (seqname!=NULL)
       { printf (
           "%*sW. Write a config file from the command line\n", OPC,"");
         printf ("%*susing the current options (for future use).\n", SPC,""); }
      printf ("%*sX. or Q. Exit program.\n", OPC,"");
      printf ("\n Enter option...");
      switch(get2char()) {
        case 'A':  Enter_seqnames();  break;
        case 'B':  Enter_opts();  break;
        case 'F':  Enter_filepath();  break;
        case 'O':  Enter_outpath();  break;
        case 'H':  Opt_help();  break;
        case 'Z':  if (ncurr>0) Clear_options();  break;
        case 'R':  Run_Gconv();  break;
        case 'S':  Show_commline();  break;
        case 'W':  Write_config();  break;
        case 'X':  case 'Q':  ex_out(); }}}



static void ex_out(void)
 { write_ini();  free_all();
   exit(0); }



static void free_all(void)
 { int i;
   static char **freelist[] =
    { &prog_name, &newpath, &inpath, &ini_name, &seqnbuff,
        &filepath, &outpath, &optbuff, NULL, NULL };
   for (i=0; freelist[i]!=NULL; i++)
     if (*freelist[i])  free(*freelist[i]);
   kill_strlist(&vars);
   kill_strlist(&opt);  kill_strlist(&sql);
   if (seqname) free(seqname);
   if (outname) free(outname);
   free(mainprog_execname);
     }




static void change_list (PSTRLIST sp)
 { char fc, ibuff[LSIZE+8];  int numadd;
   /* Ask for entry */
   printf("   ");
   if ( (fc=getline(ibuff,LBUFF))==0 || !strmatch(ibuff,"ret"))  return;
   /* Check for commands */
   if (!stricmp(ibuff,"exit"))  ex_out();
   if (!strcmp(ibuff,"%"))
    { if (sp->list==NULL)  printf ("Illegal entry!\n");
      else { kill_strlist(sp);  set_names();  newdefs=TRUE; }}
   else if (fc=='!')
    { char *ib=ibuff;
      while (isspace(*ib) || *ib=='!')  ib++;
      if (*ib==0)  printf ("Nothing to add!\n");
      else { parse_add(sp,ib,&numadd);
             if (numadd>0) { set_names();  newdefs=TRUE; }}}
   else
    { /* Don't destory old STRLIST unless something new */
      /*   is added. */
      STRLIST newstr = new_strlist();
      parse_add(&newstr,ibuff,&numadd);
      if (numadd>0)
       { kill_strlist(sp);  *sp=newstr;
         set_names();  newdefs=TRUE; }}}



static void Enter_seqnames(void)
 { /* Enter one or more input file names */
   printf (
 "\n Enter Seqfile name and/or output name and/or config files:\n"
 "    Enter `name name2' to use `name' for the sequence file\n"
 "      and `name2' for the output file name stem\n"
 "      (%s adds extensions to output file names)\n"
 "    EXAMPLES:\n"
 "      myseqs.bin\n"
 "      myseqs.bin  myseqsRun1  myconfigs.cfg\n"
 "      myseqs.bin  /g1  -nolog  -tabfileout\n"
 "    Options can be entered either here or later.\n",
         MainProgName);
   say_opts();
   /* Was there a string from before or from the INI file? */
   if (sql.n>0)  sdtoklist(sql);
   else printf (" No current entry\n");
   printf (" Enter names and options:\n");
   change_list(&sql); }



static void say_opts(void)
 { printf (
 "    Enter `Enter' only to return without changes,\n"
 "      `exit' to exit the program,\n"
 "      `%%' to REMOVE the current entry,\n"
 "      `! (options)' to ADD TO the current entry.\n"); }


static void say_opts_out(void)
 { printf (
 "    Enter `Enter' only to return without changes,\n"
 "      `exit' to exit the program,\n"
 "      `%%' to REMOVE the current entry.\n"); }



static int need_quotes(const char *str)
 { const char *pc;
   for (pc=str; *pc!=0; pc++)
     if (*pc=='"' || isspace(*pc))  return TRUE;
   return FALSE; }

static char *do_quotes(const char *str)
 { char *p, *buff=mxalloc(strlen(str)+5);
   const char *pc;
   p=buff;  *p++='"';
   for (pc=str; *pc!=0; pc++)
     if (*pc!='"')  *p++=*pc;  *p++='"';  *p=0;
   return buff; }



static STRLIST new_strlist(void)
 { STRLIST aa;  memset(&aa,0,sizeof(aa));
   return aa; }      


static void kill_strlist(PSTRLIST sp)
 { if (sp->list!=NULL && sp->n>0)
    { int n=sp->n;  char **list=sp->list;
      while(--n>=0) if (list[n])  free(list[n]);
      free(list); }
   memset(sp,0,sizeof(*sp)); }


static char *addstr (PSTRLIST sp, const char *str)
 { char *astr;
   if (need_quotes(str))  astr=do_quotes(str);
   else astr=strxdup(str);
   if (sp->n>=sp->nx)
     sp->list=rexxalloc(sp->list,sizeof(char *), sp->n,(sp->nx)+=30);
   return  sp->list[sp->n++]=astr; }



static char *pack_strlist(STRLIST s)
 { char *buff;  int i,len=0;
   for (i=0; i<s.n; i++)  len+=(strlen(s.list[i])+2);
   buff=mxalloc(len+1);
   if (s.n>0)
    { int i, len=sprintf(buff, "%s", s.list[0]);
      for (i=1; i<s.n; i++)  len+=sprintf(buff+len, "  %s", s.list[i]); }
   return buff; }



static void Enter_opts(void)
 { printf ("\n Enter options:\n");
   printf ("   EXAMPLE:\n");
   printf ("     /b2  -gscale=7\n");
   say_opts();
   if (opt.n>0)  sdtoklist(opt);
   else printf (" No current entry\n");
   printf (" Enter options:\n");
   change_list(&opt); }



static void show_current_menu_filename(const char *fname)
 { if (fname!=NULL)
    { if (strlen(fname)>50)
       { printf ("%*sCurrent entry:\n",  SPC,"");
         printf ("%*s%s\n", SPC,"", fname); }
      else
        printf ("%*sCurrent entry:  %s\n",  SPC,"", fname); }
   else  printf ("%*sNo current entry\n",  SPC,""); }



static void show_current_fname(const char *fname)
 { const char *show_name=(fname)?fname:"NONE";
   printf ("   Current entry:%s  %s\n", 
     (strlen(show_name)>50) ? "\n ":"", show_name); }




static void Enter_filepath(void)
 { char ibuff[LSIZE+8];
 try_again:
   printf ("\n Enter an input path to search for datafiles\n"
      "  and configuration files  (Optional)\n"
      "    EXAMPLE:\n"
      "       %cmydir%cmyseqs%cthisproject\n",
         PATHSEPCHAR, PATHSEPCHAR, PATHSEPCHAR);
   say_opts();
   show_current_fname(filepath);
   printf (" Enter filepath:\n   ");
   if (getline(ibuff,LBUFF)==0)  return;
   /* Check for special commands */
   if (!stricmp(ibuff,"exit"))  ex_out();
   if (!strcmp(ibuff,"%"))
    { if (filepath==NULL)
        { printf ("Illegal entry!\n");  goto try_again; }
      else { free(filepath);  filepath=NULL;  newdefs=TRUE; }
      return; }
   /* Now read the entry */
   if (ibuff[0]=='!')
    { char *rs, *ib=ibuff;
      while(isspace(*ib) || *ib=='!')  ib++;
      if (*ib==0) return;
      addstrip(ib,TRUE);
      if (filepath!=NULL)
       { sprintf(rs=mxalloc(strlen(filepath)+strlen(ib)+3),
            "%s%c%s", filepath, DIRSEPCHAR, ib);
         free(filepath);  filepath=rs; }
      else  filepath=strxdup(ib); }
   else
    { /* Replace `filepath' */
      if (filepath)  free(filepath);
      addstrip(ibuff,TRUE);  filepath=strxdup(ibuff); }
   newdefs=TRUE; }



static void Enter_outpath(void)
 { char *p, ibuff[LSIZE+8];
 try_again:
   printf ("\n Enter a directory to write output files (Optional):\n"
      "   EXAMPLE\n"
      "     myfiles%cthisproject\n", PATHSEPCHAR);
   say_opts_out();
   show_current_fname(outpath);
   printf (" Enter options:\n   ");
   if (!getline(ibuff,LBUFF))  return;
   /* Check for special commands */
   if (!stricmp(ibuff,"exit"))  ex_out();
   if (!strcmp(ibuff,"%"))
    { if (outpath==NULL)
        { printf ("Illegal entry!\n");  goto try_again; }
      else { free(outpath);  outpath=NULL;  newdefs=TRUE; }
      return; }
   if (!strcmp(ibuff,"!"))
    { printf (" Illegal entry!\n");  goto try_again; }
   /* If a path list is given, take the first directory */
   /* DIRSEPCHAR is ; for DOS, : for UNIX */
   if ( (p=strchr(ibuff,DIRSEPCHAR))!=NULL)  *p=0;
   addstrip(ibuff,TRUE);
   if (outpath) free(outpath);
   outpath=strxdup(ibuff);  newdefs=TRUE; }


static char *strtokq (const char *str)
 { static char *p, *oldstr;  char *retp;  int inq;
   if (str!=NULL)
    { if (oldstr) free(oldstr);  p=oldstr=strxdup(str); }
   while(isspace(*p)) p++;
   if (*p==0)
    { if (oldstr) free(oldstr);
      oldstr=NULL;  return NULL; }
   retp=p;  inq=0;
   while(*p!=0 && (!isspace(*p) || (inq%2)==1))
    { if (*p++=='"') inq++; }
   while (isspace(*p)) *p++=0;
   return retp; }



static int parse_add(PSTRLIST sp, const char *sbuff, int *nokp)
 { char *p, *ibuff = strxdup(sbuff);
   int nerrs=0, numok=0;
   /* Modified `strtok' to skip over quoted spaces */
   if ( (p=strtokq(ibuff))!=NULL)
    { do { if (isextent(p,"cfg"))  sp->nc++;
           /* A name, not an option */
           else if (*p!='-' && *p!='/')
            { if (sp->nf>=3)
               { printf (" More than 3 file names entered!  "
                   "(%s ignored)\n", p);  nerrs++;  continue; }
              sp->nf++; }
           addstr(sp,p);  numok++; }
      while ( (p=strtokq(NULL))!=NULL); }
   free(ibuff);
   if (nokp!=NULL) *nokp=numok;
   return nerrs; }



static void parq2(const char *p, const char *type, char **fnamep)
 { if (isextent(p,"cfg"))
     printf (" -%s=<Configname> (%s) is illegal!\n", type, p);
   else
    { if (*fnamep)  free(*fnamep);
      *fnamep=strxdup(p); }}


static void parqname(STRLIST s)
 { int i;
   for (i=0; i<s.n; i++)
    { char *p, *str=s.list[i];
      if (!strmatch(str,"-seqfile"))
       { if ( (p=strchr(str,'='))!=NULL && *++p!=0)
           parq2(str,"seqfile",&seqname); }
      else if (!strmatch(str,"-outfile"))
       { if ( (p=strchr(str,'='))!=NULL && *++p!=0)
           parq2(str,"outfile",&outname); }
      else if (!(*str=='-' || *str=='/' || isextent(str,"cfg")))
       { if (seqname==NULL)  seqname=strxdup(str);
         else if (outname==NULL)  outname=strxdup(str); }}}


static char *gpath (const char *fname)
 { char *rname=NULL;
   PARTFN pf=partfn(fname);
   if (pf.drive[0] || pf.path[0])
     rname=makefullpath(pf);
   freepf(pf);
   return rname; }


static void set_names(void)
 { char *oldseqname=seqname, *oldoutname=outname;
   /* Reset `seqname' and `outname' */
   seqname=outname=NULL;
   parqname(sql);  parqname(opt);
   if (seqname && (oldseqname==NULL || strcmp(oldseqname,seqname)!=0))
    { PARTFN pf=partfn(seqname);
      if (pf.name[0]==0)
       { pf.name=strxdup("SEQFILE");
         free(seqname);  seqname=makefmpf(pf); }
      freepf(pf); }
   /* Update outpath */
   if (outpath==NULL || strxcmp(oldseqname,seqname)!=0
        || strxcmp(oldoutname,outname)!=0)
    { char *rs;
      /* 1st choice */
      if (outname && (rs=gpath(outname))!=NULL)
       { if (outpath) free(outpath);  outpath=rs; }
      else if (outpath==NULL)
       { /* 2st choice */
         if (seqname && (rs=gpath(seqname))!=NULL)  outpath=rs;
         /* 3rd choice */
         else if (inpath)  outpath=strxdup(inpath); }}
   if (oldoutname) free(oldoutname);
   if (oldseqname) free(oldseqname); }



/*  For a possible  second screen of options?
 "    /b<blocsize>  Assume <Seqfile> is a list of nseqs/<blocsize> groups\n"
 "         with <blocsize> sequences in each group, and look only for\n"
 "         WITHIN-GROUP fragments.  Example:  /b2\n"
*/

static void Opt_help(void)
 { printf ("\n Some common options:\n"
 "   /r   Silent nucleotide sites only, permuting codon polymorphisms.\n"
 "        Default is to autosense nucleotides/proteins and use all\n"
 "        polymorphic sites.\n"
 "   /g<num>  Scale mismatch penalties by <num>.  Example: /g1 \n"
 "        Default is that fragments have no internal mismatches (/g0).\n"
 "   /n<num>   Use <num> permutations for simulated P-values.\n"
 "        Enter /n0 to use BLAST-like P-values only. Default: /n10000 \n"
 "   /c   Sequences are complete circular chromosomes, so that\n"
 "        fragments can overlap both ends of the alignment.\n"
 "   /lb  (``List best'')  List the 8 most significant global fragments\n"
 "        in any global list, with no Pvalue restrictions. With pairwise\n"
 "        lists (/lp), the best 3 per pair. The default is to list all\n"
 "        fragments with Pvalues <=0.05 up to 2000 per list.\n"
 "   /sg  Sort global fragments alphabetically by sequence names.\n"
 "        The default is to sort by increasing P-value.\n"
 "\n"
 "  Example:   /r /g1 /lb /sg   OR   /rg1lbsg \n"
 "  All unrecognized bases in <Seqfile> are treated as indels.\n"
 "  Options can also be entered in \"verbose\" form.\n"
 "  See the documentation for more details and additional options.\n");
   presskey(); }




static char *parse_path(const char *fbuff, const char *bleep)
 { size_t lenbleep=strlen(bleep);
   char *sbuff=addstripnew(fbuff,FALSE);
   if (need_quotes(fbuff))
    { char *newbuff, *qbuff, *sptr=sbuff;
      if (!strmatch(sbuff,bleep)) sptr+=lenbleep;
      qbuff=do_quotes(sptr);
      sprintf(newbuff=mxalloc(lenbleep+strlen(qbuff)+3),
           "%s%s", bleep,qbuff);
      free(qbuff);  free(sbuff);  return newbuff; }
   else
    { if (!strmatch(sbuff,bleep))  return sbuff;
      else
       { char *retname=mxalloc(strlen(bleep)+strlen(sbuff)+5);
         sprintf(retname, "%s%s", bleep,sbuff);
         free(sbuff);  return retname; }}}



static void parse_vadd (PSTRLIST sp, STRLIST a)
 { int i;
   for (i=0; i<a.n; i++)
    { char *p=a.list[i];
      if (isextent(p,"cfg"))  sp->nc++;
      else if (*p!='-' && *p!='/')
       { if (sp->nf>=3)
          { printf (" More than 3 file names!  "
                "(%s ignored)\n", p);  continue; }
         sp->nf++; }
      addstr(sp,p); }}


static void parse_vars(void)
 { int num;
   kill_strlist(&vars);
   num = vars.nx = sql.n + opt.n + 10;
   vars.list=cxalloc(num+3, sizeof(char *));
   addstr(&vars,mainprogname);
   if (filepath)  vars.list[vars.n++]=parse_path(filepath,"-filepath=");
   parse_vadd(&vars,sql);
   parse_vadd(&vars,opt);
   if (outpath)  vars.list[vars.n++]=parse_path(outpath,"-outpath=");
     }


static void Clear_options(void)
 { kill_strlist(&sql);  kill_strlist(&opt);
   if (filepath) { free(filepath);  filepath=NULL; }
   if (outname) { free(outname);  outname=NULL; }
   if (seqname) { free(seqname);  seqname=NULL; }
   if (strxcmp(inpath,outpath))
    { if (outpath) free(outpath);
      outpath=(inpath)?strxdup(inpath):NULL; }
   newdefs=TRUE; }



static char *warn_nofc(void)
 { return "WARNING: No Seqfile or configuration names entered."; }

static void disp_commline(void)
 { if (vars.n>1)
    { int col;
      printf ("\n %s command line:\n\n", MainProgName);
      col=printf(" ");
      utoklist(vars.list,vars.n, col,72, 2);
      if (seqname==NULL && vars.nf==0 && vars.nc==0)
        printf ("\n %s\n", warn_nofc()); }
   else
     printf ("\n No %s command line so far.\n\n", MainProgName); }


static void Show_commline(void)
 { /* Put the command-line options in one list */
   parse_vars();  disp_commline();
   presskey(); }



static void Write_config (void)
 { PARTFN pf,spf;  char *outcfgname;
   parse_vars();
   pf=spf=partfn(seqname);
   pf.ext=".cfg";  if (pf.name[0]==0) pf.name="config";
   outcfgname=makefmpf(pf);  freepf(spf);
   if (outpath)
    { char *s1=outcfgname;
      outcfgname=makefnam(outpath,s1);  free(s1); } 
   disp_commline();
   printf ("\n Write a configuration file for this command line?\n"
     "   Then `%s <ConfigFileName>' will have exactly the same effect\n"
     "   as running %s with these command-line options.\n",
       mainprogname, mainprogname);
   if (askd("","YN")=='Y')
    { char retchar, ibuff[LSIZE+8];  FILE *wff;
      printf ("\n Enter ConfigFileName "
        "(default: %s, `%%' to return):\n  ", outcfgname);
      if ( (retchar=getline(ibuff,LBUFF))!=0)
       { if (retchar=='%' || !stricmp(ibuff,"exit"))  return;
         if (retchar=='!') { printf ("Illegal entry!\n");  return; }
         if (!isextent(ibuff,"cfg"))
          { char *s1=outcfgname;  outcfgname=add_extent(s1,".cfg"); 
            free(s1); }}
      /* Now write the file */
      wff=fxopen(outcfgname,"wt");
      fprintf(wff, "%s\n  ", CONFIG_MAGIC);
      ftoklist(wff, vars.list+1,vars.n-1, 2,72, 2);
      fxclose(wff,outcfgname);
      printf ("\n File written: `%s'\n", outcfgname); }}



static void Run_Gconv(void)
 { char *command;
   int i, retcode;  size_t len;
   parse_vars();  disp_commline();
   if (seqname==NULL && vars.nf==0 && vars.nc==0)
    { printf (" CANNOT RUN %s without a <Seqfile> name\n"
        "  or a configuration file name.\n", MainProgName);
      presskey();  return; }

   printf("\n Run %s with these arguments?\n", MainProgName);
   if (askd("","YN")=='N')  return;

   /* This looks for `mainprogname.exe' in the default directory */
   /*   and also along the command path */
   if (searchexe(mainprog_execname))
    { printf("\n CANNOT FIND the program `%s' !!\n", mainprogname);
      printf (" In order for %s to work, the program `%s'\n"
        "   must be EITHER in the default directory "
        "OR on the command path !!\n",
          prog_name, mainprogname);
      presskey();
      return; }

   /* Disp_commline() is a `toklist' display and does not */
   /*   produce a string */
   len=strlen(mainprogname);
   for (i=1; i<vars.n; i++)  len+=(strlen(vars.list[i])+3);
   command=cxalloc(len+50,sizeof(char));
   strcpy(command,mainprogname);
   len=strlen(mainprogname);
   for (i=1; i<vars.n; i++)
     len+=sprintf(command+len, " %s", vars.list[i]);

   printf ("\n Command: %s\n\n", command);
   if ( (retcode=system(command))!=0)
     printf ("\n%s: System function returned error code %d\n",
       prog_name, retcode);

   printf ("\n Run %s on another <Seqfile> "
     "or with another set of options?\n", MainProgName);
   if (ask("","YN")=='N')  ex_out();
     }




/* Make sure that the argv[0] path is in the PATH variable */

static void setpath(const char *inpath)
 { char *p,*q, *pathvar, *fpath=strxdup(inpath);
   q=fpath;  if (fpath[0] && fpath[1]==':') q+=2;
   p = q+strlen(q);
   while (p>q && (*--p=='/' || *p=='\\')) *p=0;

   /* If no path */
   if ( (pathvar=getenv("PATH"))==NULL || pathvar[0]==0)
     sprintf(newpath=mxalloc(strlen(fpath)+15),
        "PATH=%s", fpath);
   else
     /* Prepend the current directory to the path */
     sprintf(newpath=mxalloc(strlen(fpath)+strlen(pathvar)+15),
        "PATH=%s%c%s", fpath, DIRSEPCHAR, pathvar);
   free(fpath);

   /* NOTE: The following command apparently puts a pointer */
   /*   to newpath in the environment. Therefore, `newpath' */
   /*   must NOT be freed until program exit. */
   putenv(newpath); }


static void read_ini(void)
 { char *p,*r,*s;  int i;  FILE *rff;
   if (ini_name==NULL || *ini_name==0)  return;
   if ( (rff=fopen(ini_name,"rt"))!=NULL)
    { while ( (p=lineread(rff))!=NULL) if ( (r=strchr(p,'='))!=NULL)
       { *r++=0;
         while(isspace(*r)) r++;
         s=r+strlen(r);
         while (s>r && isspace(*--s)) *s=0;
         for (i=0; iniflag[i].tag!=NULL; i++)
           if (!strmatch(p,iniflag[i].tag))
            { if (iniflag[i].initype==IPATH)
                *(iniflag[i].vname)=addstripnew(r,TRUE);
              else  *(iniflag[i].vname)=strxdup(r);
              break; }}}
   fxclose(rff,ini_name); }



static void write_ini(void)
 { FILE *wff;  int i;
   if (!newdefs) return;
   if (ini_name==NULL || *ini_name==0)  return;
   if ( (wff=fopen(ini_name,"wt"))==NULL)
    { printf ("\nCANNOT OPEN the file  %s\n", ini_name);  return; }
   fprintf(wff, "[%s]\n", prog_name);
   for (i=0; iniflag[i].tag!=NULL; i++)
    { char *p, *vv=NULL, *cbuff=NULL;  PSTRLIST sp;
      INITYPE ityp=iniflag[i].initype;
      if (ityp==ILIST && (sp=iniflag[i].sp)!=NULL)
       { if (sp->n>0)  vv=cbuff=pack_strlist(*sp); }
      else  vv=*(iniflag[i].vname);
      if (vv==NULL)  continue;
      if (ityp==IPATH)
       { while (vv[0]=='-' && (p=strchr(vv,'='))!=NULL)
          { vv=p+1;  if (*vv==0)  continue; }}
      fprintf(wff, "%s=%s\n", iniflag[i].tag, vv);
      if (cbuff!=NULL) free(cbuff); }
   fxclose(wff,ini_name); }



static char *lineread(FILE *rff)
 { static char rbuff[LSIZE+5];
   char *rb,*p;
   if (rff==NULL)  return NULL;
   do { memset(rbuff,0,sizeof(rbuff));
        if (fgets(rbuff,LSIZE,rff)==NULL)  return NULL;
        rb=rbuff;  while(isspace(*rb)) rb++;  linenum++;
        p=rb+strlen(rb);  while(p>rb && isspace(*--p))  *p=0; }
   while (*rb==0 || strchr(COMMCHARS,*rb)!=NULL);
   return rb; }



/* Utilities */

/* All-purpose file operation error messages */

static int errmsg (const char *bleep, const char *file)
 { printf ("\n%s %s the file  `%s'\n",
        (isupper(bleep[0])) ? "CANNOT" : "Cannot", bleep, file);
   perror("");  return -1; }

static void fatal (const char *bleep, const char *file)
 { errmsg (bleep, file);  exit(1); }

/* Open, close, delete files with error handling */

FILE *fxopen(const char *path, const char *mode)
 { FILE *fptr;
   if ( (fptr=fopen(path,mode))==NULL)  fatal("OPEN", path);
   return fptr; }

static int fxclose(FILE *fp, const char *fname)
 { int ret=-1;
   if (fp!=NULL && (ret=fclose(fp))!=0)  errmsg("CLOSE", fname);
   return ret; }


static int searchexe(const char *fname)
 { char *p, *spath, *sbuff;  FILE *rf;
   if (fname==NULL || fname[0]==0)  return RET_ERROR;
   /* First, check the default directory */
   if ( (rf=fopen(fname,"rb"))!=NULL)
    { fclose(rf);  return RET_OK; }
   /* Second, look for a `path' environmental variable */
   if ( (spath=getenv("path"))==NULL || spath[0]==0)
    { printf ("No program %s and no path found !\n", fname);
      return RET_ERROR; }

   /* Make a copy of the path and allocate a buffer */
   spath=strxdup(spath);
   sbuff=mxalloc((strlen(spath)+strlen(fname)+22)*sizeof(char));
   if ( (p=strtok(spath,DIRSEPSTRING))!=NULL)
     do { sprintf(sbuff, "%s%c%s", p,PATHSEPCHAR,fname);
          printf ("  LOOKING FOR %s\n",sbuff);
          if ( (rf=fopen(sbuff,"rb"))!=NULL)
           { fclose(rf);  free(sbuff); free(spath);
             return RET_OK; }}
     while ( (p=strtok(NULL,DIRSEPSTRING))!=NULL);
   free(sbuff);  free(spath);
   return RET_ERROR; }




static int getline (char *buff, int n)
 { char *p;  int len;
   fflush(stdin);
   if (n<3 || fgets(buff,n-1,stdin)==NULL)  return 0;
   fflush(stdin);
   buff[n-1]=0;  len=strlen(buff);  p=buff+len;
   while(len>0 && isspace(*--p))  { *p=0; len--; }
   if (isspace(buff[0]))
    { p=buff;  while(isspace(*p))  { p++;  len--; }
      if (len>0) memmove(buff,p,len); }
   if (len>=0 && len<n)  memset(buff+len, 0, n-len);
   return buff[0]; }



static void nl(void)  { printf("\n"); }

static int get1char (void)
 { int c;
   fflush(stdin);
   do  if ( (c=getchar())==3)  { nl();  exit(0); }
   while (c==' ' || c=='\t');   /* Ignore spaces */
   fflush(stdin);
   return c; }

static int get2char (void)
 { int c;
   return ( (c=get1char())<128)?toupper(c):c; }

static void presskey (void)
  { printf ("\n Press Enter to continue....");
    get2char(); }


/* Allocation functions with null tests and counts */

static void no_alloc (unsigned long uu, char *func)
 { printf(
    "\nCANNOT ALLOCATE %lu bytes (`%s') from the system.\n",
       uu, func);  exit(1); }


static void *cxalloc (size_t nitems, size_t size)
 { void *vptr;
   unsigned long tsize = (unsigned long)nitems*size;
   if (tsize==0L)  tsize=nitems=size=1;
   if ( (vptr = calloc(nitems,size)) == NULL)
        no_alloc (tsize, "cxalloc");
   return  vptr; }

static void *mxalloc (size_t size)
 { void *vptr;
   if (size==0)  size=1;
   if ( (vptr = calloc(1,size)) == NULL)
        no_alloc (size, "mxalloc");
   return  vptr; }

static void *rexxalloc(void *p, size_t ssize, size_t oldnum, size_t newnum)
 { void *voidptr;
   size_t newsize=ssize*(newnum+3);
   if ( (voidptr = realloc(p,newsize)) == NULL)
        no_alloc (newsize,"rexxalloc");
   if (oldnum<newnum)
     memset((char *)voidptr + ssize*oldnum, 0, ssize*(newnum+3-oldnum));
   else
     memset((char *)voidptr + ssize*newnum, 0, ssize*3);
   return voidptr; }



static char *strxdup(const char *p)
 { return (p!=NULL) ? strcpy (mxalloc(strlen(p)+1), p)
            : mxalloc(1); }


static char *strxndup(const char *p, size_t n)
 { return (p!=NULL) ? strcpy (mxalloc(strlen(p)+1+n), p)
            : mxalloc(1+n); }


static int strincmp(const char *p, const char *q, size_t n)
 { if (p==NULL) p="";  if (q==NULL) q="";
   while (n-- >0 && (*p || *q))
    { if (tolower(*p)!=tolower(*q))
         return (tolower(*p)-tolower(*q));  p++; q++; }
   return 0; }


static int strmatch(const char *str, const char *mat)
 { return strincmp((str!=NULL)?str:"",mat,strlen(mat)); }

static int stricmp(const char *p, const char *q)
 { if (p==NULL) p="";  if (q==NULL) q="";
   while (*p || *q)
    { if (tolower(*p)!=tolower(*q))
         return (tolower(*p)-tolower(*q));  p++; q++; }
   return 0; }


static int strxcmp(const char *p, const char *q)
 { if (p==NULL) p="";  if (q==NULL) q="";
   return strcmp(p,q); }



static char *makeuc(char *str)
 { char *p;
   if ( (p=str)!=NULL)
     while (*p) { *p = toupper(*p);  p++; }
   return str; }



static PARTFN partfn (const char *source)
 { int n;  char *p;  PARTFN pf;
   memset(&pf, 0, sizeof(pf));
   if (source==NULL)  source="";
   /* Is there a drive? */
   if (source[0] && source[1]==':')
    { sprintf(pf.drive=mxalloc(3), "%c:", source[0]);
      source += 2; }  else  pf.drive="";
   /* Is there a path? */
   for (n=strlen(source); --n>=0; )
     if (source[n]=='\\' || source[n]=='/')  break;
   if (++n>0)
    { memmove(pf.path = mxalloc(n+1), source, n);
      source += n; }  else  pf.path="";
   /* Is there a file name? */
   if (source[0] && source[0]!='.')
    { if ( (p=strrchr(source,'.'))!=NULL)
       { memmove(pf.name = mxalloc(p-source+1), source, p-source);
         source=p; }
      else { pf.name = strxdup(source);  source=""; }}
   else  pf.name="";
   /* Is there an extension? */
   pf.ext = (source[0]) ? strxdup(source) : "";
   return  pf; }


static char *makefmpf (PARTFN pf)
 { char *name;
   sprintf (name=mxalloc(strlen(pf.drive)+strlen(pf.path)
                +strlen(pf.name)+strlen(pf.ext)+2),
      "%s%s%s%s", pf.drive, pf.path, pf.name, pf.ext);
   return name; }


static char *makefullpath (PARTFN pf)
 { char *name;
   sprintf (name=mxalloc(strlen(pf.drive)+strlen(pf.path)+5),
      "%s%s", pf.drive, pf.path);
   return  addstrip(name,TRUE); }


static char *makefnam (const char *rawfn, const char *tplate)
 { char *mname;  PARTFN cpf,tpf, mpf;
   mpf=cpf=partfn(rawfn);  tpf=partfn(tplate);
   if (mpf.drive[0]==0)  mpf.drive=tpf.drive;
   if (mpf.path[0]==0)  mpf.path=tpf.path;
   if (mpf.name[0]==0)  mpf.name=tpf.name;
   if (mpf.ext[0]==0)  mpf.ext=tpf.ext;
   mname=makefmpf(mpf);  freepf(tpf);  freepf(cpf);
   return mname; }


static void freepf (PARTFN pf)
 { if (pf.drive[0])  free(pf.drive);
   if (pf.path[0])  free(pf.path);
   if (pf.name[0])  free(pf.name);
   if (pf.ext[0])  free(pf.ext); }


static char *addstrip(char *name, int addsep)
 { char *p=name+strlen(name);
 #ifdef WIN32_DOS
   while (p>name && (*--p=='\\' || *p=='/'))  *p=0;
   if (addsep || (name[1]==':' && name[2]==0))
    { p=name+strlen(name);  *p++=PATHSEPCHAR;  *p=0; }
 #else
   while (p>name && *--p==PATHSEPCHAR)  *p=0;
   if (addsep)
    { p=name+strlen(name);  *p++=PATHSEPCHAR;  *p=0; }
 #endif
   return name; }



static char *addstripnew(const char *str1, int addsep)
 { return addstrip(strxndup(str1,3), addsep); }


/* Return TRUE if filename=`p' has extension `ext' */
/* Otherwise return FALSE. */

static int isextent(const char *p, const char *ext)
 { const char *q;
   if (p!=NULL)
    { if ( (q=strrchr(p,'.'))!=NULL)
        return (ext!=NULL && !stricmp(q+1,ext));
      else  return (ext==NULL); }
   return FALSE; }


/* Add the extension `ext' to filename `fname' */

static char *add_extent(const char *fname, const char *ext)
 { char *newname;
   if (fname==NULL) fname="";  if (ext==NULL) ext="";
   sprintf (newname=mxalloc(strlen(fname)+strlen(ext)+3),
      "%s%s", fname, ext);
   return newname; }



static int utoklist (char **vars, int nv, int col, int collim, int rgap)
 { int ngap=0, ntot=0;
   while(--nv>=0)
    { int n=strlen(*vars);
      if ( (col+=(ngap+n)) > collim)
        { ntot+=printf("\n  ");  ngap=0;  col=n+2; }
      ntot+=printf ("%*s%s", ngap,"", *vars++);  ngap=rgap; }
   ntot+=printf("\n");
   return ntot; }


static int stoklist (STRLIST s, int col)
 { return utoklist(s.list, s.n, col, 50,1); }

static int sdtoklist (STRLIST s)
 { int col=printf (" Current entry:  ");
   return stoklist(s,col); }


static int ftoklist (FILE *wf, char **vars, int nv,
   int col, int collim, int rgap)
 { int ngap=0, ntot=0;
   while(--nv>=0)
    { int n=strlen(*vars);
      if ( (col+=(ngap+n)) > collim)
        { ntot+=fprintf(wf,"\n  ");  ngap=0;  col=n+2; }
      ntot+=fprintf(wf,"%*s%s", ngap,"", *vars++);  ngap=rgap; }
   ntot+=fprintf(wf,"\n");
   return ntot; }



static int askgen (const char *prompt, const char *anslist, int flag)
  { const char  *p;  int  c, d;
    if (anslist==NULL || *anslist == 0)  return '?';
    do { p = anslist-1;
         fprintf(stderr, "%s  Enter `%c'", prompt, *++p);
         while ( (d=*++p)!=0)  fprintf(stderr, " or `%c'", d);
         if (flag)  fprintf(stderr, "  (default `%c')", *--p);
         fprintf(stderr, "....");  c = get2char();
         p = anslist-1;
         while ( (d=toupper(*++p))!=0 &&  d != c)  ; }
    while (d == 0 && !flag);
    return  (d) ? *p : *--p; }


static int ask(const char *prompt, const char* anslist)
 { return askgen(prompt,anslist,0); }

static int askd(const char *prompt, const char* anslist)
 { return askgen(prompt,anslist,1); }


