
/* Functions using writing ``dump'' output files */

/* Parts of source file (line numbers approximate): */

/* (***AAA) (L24) Main output file routines */

/* (***BBB) (L1317) Jseq and plotfile routines */

/* (***CCC) (L1764) Utilities used in output (dump) routines: */
/*     show_seqnames(), show_polysites(), show_mainpvals(), etc. */


#include "vall.h"
#include "vflags.h"
#include "vread.h"
#include "vcalc.h"
#include "vdump.h"

char *AllRightsReserved_Vdump =
   "All rights reserved - Stanley Sawyer";


/* (***AAA) Functions for writing an ASCII flat file */
/*   with one or more lists of significant fragments */

/* Routines for writing .frag, .tab, .dif, and .jseq files */


#ifdef VTEXT
static void list_gfrags(SIMX *sc);
static void list_gjseqs(SIMX *sc);
static void dump_maxmeansims(void);
#define VMFRAG  MFRAG

#else /* If VSHOW */
#define VMFRAG  MFRAGPV

#endif /* VTEXT */


static int dump_fragsnorm (char *dfile, DX_TYPE dx_type);
static int dump_jseqs(char *ffile);
static int dump_plot(char *ffile);
static void write_dump_heading(char *bleep, DX_TYPE dx_type);
static void frag_summary_heading(void);
static void set_annotate(const char *dfile);
static void say_indel_aligned_offsets(void);
static void say_written (const char *bleep, const char *fnlist);
static void say_nofrags(const char *bleep, int extraflag);
static void show_mainpvals (DX_TYPE dx_type);
static void list_pfrags(SIMX *sc);
static void list_pairlist (SIMTYPE simtype, int vflags,
     PAIRCOMP *pc, int *pgroup, int npairs, int mxlen);

#ifdef VTEXT
static char *gnamelist(int allflag, int bleepflag);
#endif
static char *pnamelist(int allflag, int bleepflag);

static void show_seq_path_lists(void);
static void set_tnlisted(void);
static void clear_tnlisted(void);

static void fragtype_linedescr (SIMTYPE simtype,
    int (*prfun)(const char *control, ...));
static void set_ftag (char *ftag, char tchar, SIMTYPE simtype);
static void Warn_nofullseqs(const char *dfile, int annotflag);

static int colwid (SIMTYPE simtype, int i1, int i2, int g);

static int need_unal, *unal_len, **unal_help;
static int dokapairpvals, dokaglobpvals;
static char *dump_knownchars;


/* Functions and routines for writing to a dump file */

static FILE *dpff;

static int tab_dumpnumform(const char *control, ...);
static int tab_dumpnum(int num);

/* These functions write a single field to a tab or frags file. */
/* `Dumpnl' and `dumptext' are the generic functions. */
/* `Dumpxnl' and `dumpxtext' are designed for the first */
/*   field on a line (or for an entire line) and generally */
/*   prepend '# '  */
/* In tab format, 'dumptext' and 'dumpxtext' postpend tab characters */
/*   UNLESS the string ends in one or more newline chars. */
/* NOTE: '# ' is NOT prepended to succeedings lines if */
/*   'dumptext' or 'dumpxtext' strings have internal newlines */
/*   or more than one final newline characters. */
/*   Thus, currently, these functions should not be used */
/*   for multiline text, although it would be easy to rewrite */
/*   them so that they did prepend '# ' on succeeding lines. */

/* DIF format spreadsheet files seem to require a special format */
/*   for numerical fields.  'Dumpnum' and 'dumpnumform' write */
/*   numbers WITHOUT a prepended '# ' to a field.  For tab files, */
/*   `dumpnum' and `dumpnumform' always postpend a tab character. */

static void (*dumpnl)(void) = BNL;
static void (*dumpxnl)(void) = BNL;
static int (*dumptext)(const char *control, ...) = BPRINTF;
static int (*dumpxtext)(const char *control, ...) = BPRINTF;
static int (*dumpnumform)(const char *control, ...) = tab_dumpnumform;
static int (*dumpnum)(int num) = tab_dumpnum;

static int taboutflag, no_onlylist;

#define DBUFFSIZE 2000   /* Line buffer size */

/* Functions for output files for dx_type==DX_FRAGS or DX_JSEQS etc */

static void norm_dumpnl(void)
 { fprintf(dpff, "\n"); }

static void norm_dumpxnl(void)
 { fprintf(dpff, "#\n"); }

static int norm_dumptext(const char *control, ...)
 { int ret;  va_list list;  va_start(list, control);
   ret=vfprintf(dpff, control, list);
   va_end(list);
   return ret; }

static int norm_dumpxtext(const char *control, ...)
 { int ret=fprintf (dpff, "# ");
   va_list list;  va_start(list, control);
   ret+=vfprintf(dpff, control, list);
   va_end(list);
   return ret; }


/* Functions for dx_type==DX_DIF */

/* This format requires that all fields be enclosed in a pair */
/*   of "s unless the field is purely numerical */

static void dif_dumpnl(void)
 { fprintf (dpff, "-1,0\nBOT\n"); }

static int dif_dumptext(const char *control, ...)
 { int ret,nret;  char *p, nbuff[DBUFFSIZE];
   va_list list;  va_start(list, control);
   vsprintf(nbuff, control, list);
   /* Remove and count trailing newline chars: */
   p=nbuff+strlen(nbuff);  nret=0;
   while (p>nbuff && *--p=='\n')  { *p=0;  nret++; }
   ret=fprintf (dpff, "1,0\n\"%s\"\n", nbuff);
   if (nret>0)  while (--nret>=0) { dif_dumpnl();  ret++; }
   va_end(list);
   return ret; }

static int dif_dumpxtext(const char *control, ...)
 { int ret, nret;  char *p, nbuff[DBUFFSIZE];
   va_list list;  va_start(list, control);
   vsprintf(nbuff, control, list);
   /* Remove and count trailing newline chars: */
   p=nbuff+strlen(nbuff);  nret=0;
   while (p>nbuff && *--p=='\n')  { *p=0;  nret++; }
   ret=fprintf (dpff, "1,0\n\"# %s\"\n", nbuff);
   if (nret>0)  while (--nret>=0) { dif_dumpnl();  ret++; }
   va_end(list);
   return ret; }


static int dif_dumpnum(int num)
 { return  fprintf (dpff, "0,%d\nV\n", num); }

static int dif_dumpnumform(const char *control, ...)
 { int ret;  char nbuff[DBUFFSIZE];
   va_list list;  va_start(list, control);
   vsprintf(nbuff, control, list);
   ret=fprintf (dpff, "0,%s\nV\n", nbuff);
   va_end(list);
   return ret; }


/* Functions for dx_type==DX_TAB */

/* If the string contains commas but no quotes, */
/*   enclosed in quotes except for trailing newline characters. */
/* Add a tab character if the string does not end with an EOL */
/* This functions assumes buf-1 is valid and that there is */
/*   some additional room at the end of buf. */

static char *tab_enc (char *buf)
 { char *p;  int n;
   if (strchr(buf,',')!=NULL && strchr(buf,'"')==NULL)
    { /* Remove and count trailing newline chars: */
      char qc='"';  int nret=0;  p=buf+strlen(buf);
      while (p>buf && *--p=='\n')  { *p=0;  nret++; }
      p=buf+strlen(buf);      /* Reposition at end */
      /* Enclose buf in double quotes */
      *--buf=qc;  *p++=qc;
      /* Replace any newline characters */
      while (--nret>=0)  *p++='\n';  *p=0; }
   n=strlen(buf);
   if (n==0 || buf[n-1]!='\n')  { buf[n++]='\t';  buf[n]=0; }
   return buf; }


static int tab_dumptext(const char *control, ...)
 { int ret;  char *buf, nbuff[DBUFFSIZE];
   va_list list;  va_start(list, control);
   buf=nbuff+5;  vsprintf(buf, control, list);
   ret=fprintf (dpff, "%s", tab_enc(buf));
   va_end(list);
   return ret; }

static int tab_dumpxtext(const char *control, ...)
 { int ret;  char *buf, nbuff[DBUFFSIZE];
   va_list list;  va_start(list, control);
   buf=nbuff+5;  strcpy(buf,"# ");
   vsprintf(buf+2, control, list);
   ret=fprintf (dpff, "%s", tab_enc(buf));
   va_end(list);
   return ret; }


static int tab_dumpnumform(const char *control, ...)
 { int ret;  va_list list;  va_start(list, control);
   ret=vfprintf(dpff, control, list);
   ret+=fprintf (dpff, "\t");
   va_end(list);
   return ret; }

static int tab_dumpnum(int num)
 { return  fprintf (dpff, "%d\t", num); }



/* Function and routines for opening a dump file */
/*   and initializing write functions */

/* NOTE: Empty lines look better with an initial # */
/*   in DX_FRAGS format, but not in spreadsheet formats. */

static void dumpx_open(char *dfile, DX_TYPE dx_type)
 { /* Open the dump file */
   if (dfile!=NULL)
    { dpff=fxopenwrite(dfile,WRITEMODE);  dumpflag=TRUE; }
   else  dx_type=DX_NONE;
   /* Set or reset write functions */
   switch(dx_type) {
     default:  textprintf=dumptext=dumpxtext=BPRINTF;
        textnl=dumpnl=dumpxnl=BNL;  dumpflag=taboutflag=FALSE;
        dumpnumform=tab_dumpnumform;  dumpnum=tab_dumpnum;  break;
     case DX_FRAGS:  case DX_JSEQS:  case DX_PLOT:
        textnl=dumpnl=norm_dumpnl;  dumpxnl=norm_dumpxnl;
        dumptext=norm_dumptext;  taboutflag=FALSE;
        textprintf=dumpxtext=norm_dumpxtext;  break;
     case DX_DIF:
        textnl=dumpnl=dumpxnl=dif_dumpnl;
        dumptext=dif_dumptext;  taboutflag=TRUE;
        textprintf=dumpxtext=dif_dumpxtext;
        dumpnumform=dif_dumpnumform;  dumpnum=dif_dumpnum;  break;
     case DX_TAB:
        textnl=dumpnl=dumpxnl=norm_dumpnl;
        dumptext=tab_dumptext;  taboutflag=TRUE;
        textprintf=dumpxtext=tab_dumpxtext;
        dumpnumform=tab_dumpnumform;  dumpnum=tab_dumpnum;  break; }}



void dumpx_close(char *dfile)
 { if (dpff) { fxclose(dpff,dfile);  dpff=NULL; }
   dumpx_open(NULL,DX_NONE); }


static void free_unal(void)
 { if (unal_len!=NULL) { free(unal_len);  unal_len=NULL; }
   if (unal_help!=NULL)
    { int n=nseqs;
      while (--n>=0)  free(unal_help[n]);
      free(unal_help);  unal_help=NULL; }}


/* Make sure that these files can be opened !! */
void testdump(void)
 { int i;  FILE *wff;
   DUMP_STRUCT *dds=parse_dfiles(outfile);
   for (i=0; dds[i].dfile!=NULL; i++)
    { wff=fxopenwrite(dds[i].dfile,WRITEMODE);
      fxclose(wff,dds[i].dfile); }}


/* Write all relevant dumpfiles */
/* Dump_fragsnorm, dump_jseqs(), and dump_plot() */
/*   are called ONLY from dump_fragments() */

int dump_fragments(DUMP_STRUCT *dds)
 { char *flist[10];
   int i,num, locflag=FALSE;  DX_TYPE dx;
   PRINTFUNP oldprfun;  void (*oldprnl)(void);
   if (dds==NULL)
    { dds=parse_dfiles(outfile);  locflag=TRUE; }
   set_totpairparms();  need_unal=TRUE;
   #ifdef VTEXT
   set_totglobparms();
   #endif
   no_onlylist=(is_onlylist()==FALSE);     /* Check to be sure */

   /* DON'T COUNT IUB ambiguous-nucleotide characters as indels */
   /*   for the purpose of computing TRUE offsets: */
   /* Set_offtypes() returns amino acids or IUB nucleotides */
   dump_knownchars=set_offtypes();
   /* Write the output files in the list */
   for (num=i=0; dds[i].dfile && i<8; i++)
    { int x;
      BPRINTF ("Written to %s:\n", dds[i].dfile);
      switch (dx=dds[i].dx_type) {
        default:  x=dump_fragsnorm(dds[i].dfile, dx);  break;
        case DX_JSEQS:  x=dump_jseqs(dds[i].dfile);  break;
        case DX_PLOT:   x=dump_plot(dds[i].dfile);  break; }
      if (x!=0)
       { char *fn=flist[num++]=mxalloc(strlen(dds[i].dfile)+4);
         sprintf(fn,"`%s'", dds[i].dfile); }
      if (locflag) free(dds[i].dfile); }

   oldprfun=textprintf;  oldprnl=textnl;
   reset_textprintf();
   ftoklistf("OUTPUT FILES WRITTEN:  ", flist,num, 1,NULL,0, TOK_FREE);
   if (havelog) textprintf ("LOG FILE:  `%s'\n", logfile);
   else textprintf ("No log file written.\n");
   set_textprintf(oldprfun,oldprnl);
   /* Restrict lists for first set of output files only */
   kill_onlylist(FALSE);
   free_unal();
   return num; }



/* Create a list of dumpfile names and types */
/* `dds' should point to at least 5 structures */

char *outfile_ext_list[] =
  { "frags", "dif", "tab", "jseqs", "plot", "sum", NULL, NULL };

typedef struct { int *dflag;
                 char *ext;  DX_TYPE dx_type; }  PARSE_DSTRUCT;

DUMP_STRUCT *parse_dfiles(char *outname)
 { int i,n, fflag=FALSE;
   static DUMP_STRUCT ddb[10];
   static PARSE_DSTRUCT pdstruct[] = {
     { &dumpfrags, ".frags", DX_FRAGS },
     { &dumptab,   ".tab",   DX_TAB },
     { &dumpdif,   ".dif",   DX_DIF },
     { &dumpjseqs, ".jseqs", DX_JSEQS },
     { &dumpplot,  ".plot",  DX_PLOT },
     { NULL }, { NULL } };
   /* Get the outfile (drive,) path, and name: */
   if (outname==NULL || outname[0]==0)  outname=seqfile;
   else { outname=makefnam(outname,seqfile);  fflag=TRUE; }
   memset(ddb,0,sizeof(ddb));
   for (n=i=0; pdstruct[i].dflag!=NULL; i++)
     if (*(pdstruct[i].dflag))
      { ddb[n].dfile=makefnam(pdstruct[i].ext,outname);
        ddb[n].dx_type=pdstruct[i].dx_type;  n++; }
   if (fflag) free(outname);
   return  ddb; }


/* Write an output FRAG, TAB, or DIF file */
/* Dump_fragsnorm is called only from dump_fragments() */

static int dump_fragsnorm (char *dfile, DX_TYPE dx_type)
 { /* Open the output file and initialize */
   /*   the file writing functions */
   dumpx_open(dfile,dx_type);
   if (dx_type==DX_DIF)
     fprintf(dpff, "TABLE\n0,1\n\"EXCEL\"\n-1,0\nBOT\n");

   /* Start with the heading */
   write_dump_heading("Output",dx_type);
   show_mainpvals(dx_type);
   #ifdef VTEXT
   if (numsims>0 && show_mmsims)
    { dumpxnl();
      if (taboutflag) dump_maxmeansims();
      else show_maxmeansims(FALSE); }
   #endif
   frag_summary_heading();
   if (annotate)  set_annotate(dfile);

   /* If VTEXT is set, display global fragments */
   clear_tnlisted();   /* Set `nlisted' counts=0 */
   #ifdef VTEXT
   if (listglobal>0)
    { dokapairpvals=showbcpwkapvals;  dokaglobpvals=TRUE;
      list_gfrags(&cfinx);
      list_gfrags(&cfoutseqx);
      list_gfrags(&cfoutpairx);
      list_gfrags(&cfoutgrpx);
      say_written("global", gnamelist(FALSE,FALSE)); }
   else  say_nofrags("global",FALSE);
   #endif

   /* List pairwise fragments */
   if (listpairs && listperpair>0)
    { dokapairpvals=TRUE;  dokaglobpvals=showbcpwkapvals;
      list_pfrags(&cfinx);
      list_pfrags(&cfoutseqx);
      list_pfrags(&cfoutpairx);
      list_pfrags(&cfoutgrpx);
      #ifdef VTEXT
      say_written("pairwise", pnamelist(FALSE,FALSE));
      #else
      say_written("", pnamelist(FALSE,FALSE));
      #endif
         }
   else  say_nofrags("pairwise",FALSE);

   dumpxnl();
   dumpx_close(dfile);
   return TRUE; }





static void write_dump_heading(char *bleep, DX_TYPE dx_type)
 { char *sqfile, nbuff[88];  int i;
   /* Write the heading and command line (including CONFIG files) */
   /* Since `dumpxtext()' may prepend `# ' to its output, */
   /*   dumpxtext() should not be used for more than one line. */
   dumpxtext("%s of %s for sequence file %s\n",
     bleep, ProgName, seqfile);
   dumpxtext ("%s version %s\n", ProgName, ProgVersion);
   kill_testlist(FALSE);    /* Don't write to DUMP files */
   show_commline(NULL,FALSE);
   dumpxnl();
   if (seqfixed_dir && seqfdir!=NULL)  sqfile=seqfdir;
   else sqfile=seqfile;
   if (seqfixed_dir || (seqfdir==NULL && filepath.s==NULL))
     dumpxtext("%s: %s%c\n", sqfile,
       numsay(nbuff,file_nseqs, "sequence",TRUE),
       (seqfcomments.s) ? ':' : '.');
   else
     dumpxtext("%s (%s): %s%c\n", sqfile,
       (seqfdir) ? seqfdir : "in default directory",
       numsay(nbuff,file_nseqs, "sequence",TRUE),
       (seqfcomments.s) ? ':' : '.');
   if (seqfcomments.s!=NULL)
     for (i=0; seqfcomments.s[i]!=NULL; i++)
       dumpxtext("  %s\n", seqfcomments.s[i]);
   dumpxnl();

   /* Show a selected list of features */
   /* (Recall that textprintf and ftoklist have been redirected.) */
   show_features();
   show_seqnames();   /* Plus some explanatory text */
   show_seq_path_lists();
   dumpxnl();
   gscale_comment();
   #ifdef VSHOW
   if (tabfileflag==FALSE)
   #endif
    { show_basedata();
      if (dx_type!=DX_JSEQS && dx_type!=DX_PLOT)
       { if (seqtype==SILENT)  show_codon_stats(); }
      if (saveseqs && showunal)
       { dumpxnl();  show_unalens(dumpxtext,dumptext); }
      if (dx_type!=DX_JSEQS && dx_type!=DX_PLOT)
       { if (showpolyflag)  show_polysites(); }}
   show_kaparm_errs(dumpxtext); }



static void frag_summary_heading(void)
 { dumpxnl();
   #ifdef VTEXT
   dumpxtext("Global lists: %s\n", gnamelist(TRUE,TRUE));
   dumpxtext("Pairwise lists: %s\n", pnamelist(TRUE,TRUE));
   if ( (listpairs && listperpair>0) || listglobal>0)
   #else
   dumpxtext ("%s\n", pnamelist(TRUE,TRUE));
   /* `listpairs' is not used if VSHOW */
   if (listperpair>0)
   #endif
    { dumpxtext(" where I means inner pair, O is outer seq, "
          "A is outer pair, and\n");
      dumpxtext(" G means outer group (fragments).\n");
      if (numsims>0)
        dumpxtext("Global P-values are %s",  (bcsims)
          ? "Bonferroni-corrected pairwise P-values.\n"
          : "based on BLAST-like global scores.\n");  }

   dumpxnl();
   if (tothavcirc && taboutflag==FALSE)
    { dumpxtext(
       "`CA' stands for `Circle Around', and means that the fragment\n");
      dumpxtext("  encircles the endpoints.\n"); }
   dumpxtext ("Fragment offsets and lengths are in %s.\n",
       (seqtype==PROTEIN) ? "codons" : "nucleotides");
   if (numdelblocs>0)
    { dumpxtext("Note: A polymorphism can be at a run "
           "of sites with indels,\n");
      dumpxtext("  which here have up to %d sites.\n",
        maxdelblocs); }
   dumpxtext("Bases other than `%s' are treated as indels\n",
     set_pchars());
   dumpxtext("  for the purpose of generating polymorphisms.\n"); }



static void set_annotate(const char *dfile)
 { if (saveseqs==FALSE || seqbase==NULL || r_polyflag)
    { Warn_nofullseqs(dfile,TRUE);
      annotate=showunal=dumpjseqs=FALSE; }
   else
    { if (need_unal)
       { int i;  free_unal();
         unal_len=cxalloc(nseqs+2,sizeof(int));
         unal_help=cxalloc(nseqs+2,sizeof(int *));
         for (i=0; i<nseqs; i++)
          { char *str=seqbase[i];
            int j,len, ulen,*uhelp;
            ulen=nbases/100;  if (nbases%100) ulen++;
            uhelp=cxalloc(ulen+3,sizeof(int));
            for (len=j=0; j<ulen; j++)
             { int k, j1=100*j;
               for (k=0; k<100 && j1<nbases; k++, j1++)
                 if (strchr(dump_knownchars,str[j1]))  len++;
               uhelp[j+1]=len; }
            unal_len[i]=len;
            unal_help[i]=uhelp; }
         need_unal=FALSE; }
      say_indel_aligned_offsets(); }}



static void say_indel_aligned_offsets(void)
 { dumpxtext("Bases other than `%s' are treated as indels\n",
      dump_knownchars);
   dumpxtext("  for the purpose of computing unaligned offsets.\n"); }


static void say_written (const char *bleep, const char *fnlist)
 { const char *vars[8];  int n;
   PRINTFUNP oldprfun=textprintf;  void (*oldprnl)(void)=textnl;
   n=0;
   vars[n++]=fnlist;
   vars[n++]="significant";
   if (bleep!=NULL && bleep[0]!=0) vars[n++]=bleep;
   vars[n++]="fragments";
   set_textprintf(BPRINTF,BNL);
   ftoklistf("  ", (char **)vars,n, 1,NULL,0, TOK_ENDPERIOD);
   set_textprintf(oldprfun,oldprnl); }


static void say_nofrags(const char *bleep, int extraflag)
 { dumpxnl();  dumpxnl();
   dumpxtext("No %s fragment lists.\n", bleep);
   dumpxtext("(See earlier in the file for the "
      "restrictions on fragments.)\n");
   if (extraflag) dumpxnl();
   BPRINTF("  No %s fragments.\n", bleep); }


static void tab_fragphead(char *bleep, char **blist, int *bp, int circ)
 { dumptext("");  dumptext("%s",bleep);  dumptext("");
   if (circ)  dumptext("Circle");
   blist[(*bp)++]="Begin";  blist[(*bp)++]="End";
   blist[(*bp)++]="Length";
   if (circ)  blist[(*bp)++]="Around?"; }

static void tab_fragpos (FRAGOFFS x, int nb, int circ)
 { int lenf =  x.rt - x.lf + 1;  if (lenf<=0)  lenf+=nb;
   dumpnum(x.lf + AddOffset);  dumpnum(x.rt + AddOffset);  dumpnum(lenf);
   if (circ) dumptext( (x.caf) ? "Yes" : ""); }

static void addl(char *buff, int off, char *str)
 { memmove(buff+off, str, strlen(str)); }

static void norm_fragphead(char *ab, char *bb, int off, char *text)
 { int n=(strlen(text)+1)/2;
   addl(ab,off+8-n,text);
   addl(bb,off,"Begin");
   addl(bb,off+7,"End");  addl(bb,off+13,"Len"); }

static char *norm_fragpos (char *bf, FRAGOFFS x, int nb)
 { int lenf =  x.rt - x.lf + 1;  if (lenf<=0)  lenf+=nb;
   sprintf(bf, "%4d %s%4d %4d", x.lf + AddOffset,
     (x.caf) ? "CA" : "  ",  x.rt + AddOffset, lenf);
   return bf; }


static int unaloff(int ll, int i)
 { int n,j,j1=(--ll)/100;  char *sq=seqbase[i];
   n=unal_help[i][j1];  j=j1*100;
   while (j<ll)  if (strchr(dump_knownchars,sq[j++]))  n++;
   return n+1; }

/* Flags for fraglist functions */

#define HD_SIMS    1
#define HD_GLOBAL  2
#define HD_BCSIM   4


static void tab_listhead(SIMTYPE simtype, int vflags)
 { char *blist[200], *pen, *totpen;  int i, nb=0;
   int dosims = (vflags & HD_SIMS), global = (vflags & HD_GLOBAL);
   int bcsim = (vflags & HD_BCSIM);
   int pairflag = (simtype==SIM_INNER || simtype==SIM_OUTERPAIR);
   for (i=0; i<200; i++)  blist[i]="";
   dumpxtext("Frag");  blist[nb++]="  type";
   dumptext( (simtype==SIM_OUTERGROUP)?"Group":"Sequence" );
   blist[nb++] = (pairflag)?"Names":"Name";
   if (dosims)
    { dumptext( (bcsim) ? "BC Sim" : "Sim");  blist[nb++]="Pvalue"; }
   if (dokapairpvals)
    { dumptext("KA");  blist[nb++]="Pvalue"; }
   if (dokaglobpvals)
    { dumptext("BC KA"); blist[nb++]="Pvalue"; }
   if (showlog10pvals)
    { dumptext("(-)Log10");
      blist[nb++]=(global)?"BCKAPv":"KAPval"; }
   tab_fragphead("Aligned Offsets",blist,&nb,circular);
   if (annotate)
    { tab_fragphead( (pairflag) ? "In Seq1" : "In Seq",
          blist, &nb, FALSE);
      if (pairflag) tab_fragphead("In Seq2", blist,&nb,FALSE); }
   if (simtype==SIM_INNER) { pen="Dif";  totpen="Diffs"; }
   else { pen="Mat";  totpen="Mats "; }
   dumptext("Num");  blist[nb++]="Poly";
   dumptext("Num");  blist[nb++]=pen;
   dumptext("Total");  blist[nb++]=totpen;
   dumptext("Mismatch");  blist[nb++]="Penalty";
   dumpnl();
   dumpxtext(blist[0]);
   for (i=1; i<nb; i++)  dumptext(blist[i]);  dumpnl(); }


static void tab_listline(char *ftag, char *names, int i1,int i2,
    double kapval, int npairs, VMFRAG *fp,
    int totdifs, int mismatpen, int vflags)
 { char *nmask;  int frpolys;
   int dosims = (vflags & HD_SIMS), global = (vflags & HD_GLOBAL);
   /* typedef struct { int lf,rt,caf; } FRAGOFFS; */
   FRAGOFFS x;  memset(&x,0,sizeof(x));
   #ifdef VSHOW
   if (tabfileflag)
    { if ( (x.rt=fp->finish) < (x.lf=fp->start) ) x.caf=TRUE;
      frpolys=-1; }
   else
   #endif
    { x = offsfrags(fp->start,fp->finish);
      frpolys = fp->finish - fp->start; }

   /* Fragment lines DO NOT BEGIN with # ! */
   dumptext(ftag);  dumptext(names);
   if (expformat)  nmask="%8.2e";
   else if (widecols)  nmask="%9.7f";
   else  nmask="%7.5f";

   if (dosims)
    { double pval;
   #ifdef VTEXT
      pval=(double)*fp->grc/numsims;
   #else
      pval=pow(10,-fp->logpvsim);
   #endif
      if (truncate_pvals && pval>1.0)  pval=1.00;
      if (pval>1.0)  dumptext("> 1.0");
      else dumpnumform ("%6.4f", pval); }
   if (dokapairpvals)  dumpnumform (nmask, kapval);
   if (dokaglobpvals)
    { double mpval=kapval*npairs;
      if (truncate_pvals && mpval>1.0)  mpval=1.00;
      if (mpval>1.0)  dumptext("> 1.0");
      else  dumpnumform (nmask, mpval); }
   if (showlog10pvals)
    { double mpval = (global) ? kapval*npairs : kapval;
      if (truncate_pvals && mpval>1.0)  mpval=1.00;
      if (mpval>1.0)  dumptext("< 0.0");
      else { char *smask=(widecols) ? "%9.5f" : "%7.3f";
             dumpnumform(smask, -log10(mpval)); }}
   tab_fragpos(x, nbases,circular);
   if (annotate)
    { FRAGOFFS xan=x;
      xan.lf=unaloff(x.lf,i1);  xan.rt=unaloff(x.rt,i1);
      tab_fragpos(xan,unal_len[i1],FALSE);
      if (i2>=0)
       { xan.lf=unaloff(x.lf,i2);  xan.rt=unaloff(x.rt,i2);
         tab_fragpos(xan,unal_len[i2],FALSE); }}
   dumpnum(frpolys);
   dumpnum(fp->nmismat);
   dumpnum(totdifs);
   if (gscale==0)  dumptext("None");
   else dumpnum(-mismatpen);
   dumpnl(); }




static void norm_listhead(int mxlen, SIMTYPE simtype, int vflags)
 { char aline[310], bline[310], *pen, *totpen;
   int dosims = (vflags & HD_SIMS), global = (vflags & HD_GLOBAL);
   int bcsim = (vflags & HD_BCSIM);
   int x,rlen,kawid;
   int pairflag=(simtype==SIM_INNER || simtype==SIM_OUTERPAIR);
   memset(aline,' ',300);  memset(bline,' ',300);
   aline[0]=bline[0]='#';
   addl(aline,4,(simtype==SIM_OUTERGROUP)?"Group":"Seq");
   addl(bline,4,(pairflag)?"Names":"Name");
   if (expformat) kawid=10;
   else if (widecols) kawid=11;  else kawid=9;
   x=mxlen+5;
   if (dosims)
    { if (bcsim)  addl(aline,x,"BC Sim");
      else  addl(aline,x+1,"Sim");  addl(bline,x,"Pvalue");  x+=8; }
   if (dokapairpvals)
    { int x2=x;  if (expformat==FALSE && widecols) x2++;
      if (dosims)  x2+=2;
      addl(aline,x2,"KA");  addl(bline,x,"Pvalue");
      x+=kawid; }
   if (dokaglobpvals)
    { int y=(expformat || widecols) ? x+1 : x;
      addl(aline,y,"BC KA");  addl(bline,y,"Pvalue");
      x+=kawid; }
   if (showlog10pvals)
    { if (widecols) x++;
      addl(aline,x-1,"-Log10");
      addl(bline,x,(global)?"BCKAPv":"KAPval");
      x+=8;  if (widecols) x++; }
   norm_fragphead(aline,bline,x,"Aligned Offsets");
   rlen=19;  x+=(rlen-1);
   if (annotate)
    { norm_fragphead(aline,bline,x+2,
          (pairflag) ? "In Seq1" : "In Seq");  x+=(2+rlen);
      if (pairflag)
       { norm_fragphead(aline,bline,x,"In Seq2");  x+=rlen; }}
   if (simtype==SIM_INNER) { pen="Dif";  totpen="Difs"; }
   else { pen="Mat";  totpen="Mats"; }
   addl(aline,x,"Num");  addl(bline,x,"Poly");  x+=5;
   addl(aline,x,"Num");  addl(bline,x,pen);  x+=5;
   addl(aline,x,"Tot");  addl(bline,x,totpen);  x+=5;
   addl(aline,x,"MisM");  addl(bline,x,"Pen.");  x+=4;
   aline[x]=bline[x]=0;
   fprintf(dpff, "%s\n%s\n", aline,bline); }



static void norm_listline(char *ftag, char *names,int mxlen,
    int i1,int i2, double kapval, int npairs, VMFRAG *fp,
    int totdifs, int mismatpen, int vflags)
 { char nbuff[88], abuff[22], pvalbuff[277], *nmask;
   int len=0, n2len, frpolys;
   int dosims = (vflags & HD_SIMS), global = (vflags & HD_GLOBAL);
   /* typedef struct { int lf,rt,caf; } FRAGOFFS; */
   FRAGOFFS x;  memset(&x,0,sizeof(x));
   #ifdef VSHOW
   if (tabfileflag)
    { if ( (x.rt=fp->finish) < (x.lf=fp->start) ) x.caf=TRUE;
      frpolys=-1; }
   else
   #endif
    { x = offsfrags(fp->start,fp->finish);
      frpolys = fp->finish - fp->start; }

   if (gscale==0)  strcpy(abuff,"None");
   else sprintf(abuff,"%3d",-mismatpen);
   if (expformat) { nmask="%8.2e";  n2len=8; }
   else if (widecols)  { nmask="%9.7f";  n2len=9; }
   else  { nmask="%7.5f";  n2len=7; }

   if (dosims)
    { double pval;
   #ifdef VTEXT
      pval=(double)*fp->grc/numsims;
   #else
      pval=pow(10,-fp->logpvsim);
   #endif
      if (truncate_pvals && pval>1.0)  pval=1.00;
      if (pval>1.0) len=sprintf(pvalbuff, "%-6s", "> 1.0");
      else len=sprintf(pvalbuff, "%6.4f", pval); }
   if (dokapairpvals)
    { if (len>0)  len+=sprintf(pvalbuff+len, "  ");
      len+=sprintf (pvalbuff+len, nmask, kapval); }
   if (dokaglobpvals)
    { double mpval;
      if (len>0)  len+=sprintf(pvalbuff+len, "  ");
      if ( (mpval=kapval*npairs)>1.0 && truncate_pvals)  mpval=1.00;
      if (mpval>1.0)
        len+=sprintf(pvalbuff+len, "%-*s", n2len, "> 1.0");
      else  len+=sprintf (pvalbuff+len, nmask, mpval); }
   if (showlog10pvals)
    { double mpval = (global) ? kapval*npairs : kapval;
      if (mpval>1.0 && truncate_pvals)  mpval=1.00;
      if (mpval>1.0)
        sprintf(pvalbuff+len, "%-*s", (widecols)?10:8, "  < 0.0");
      else
       { char *smask=(widecols) ? "%9.5f " : "%7.3f ";
         sprintf (pvalbuff+len, smask, -log10(mpval)); }}

   fprintf (dpff, "%s  %-*s %s  %s",
      ftag, mxlen,names, pvalbuff, norm_fragpos(nbuff, x, nbases));
   if (annotate)
    { FRAGOFFS xan=x;
      xan.lf=unaloff(x.lf,i1);  xan.rt=unaloff(x.rt,i1);
      fprintf(dpff, "    %s",
        norm_fragpos(nbuff, xan, unal_len[i1]));
      if (i2>=0)
       { xan.lf=unaloff(x.lf,i2);  xan.rt=unaloff(x.rt,i2);
         fprintf(dpff, "   %s",
            norm_fragpos(nbuff, xan, unal_len[i2])); }}
   fprintf (dpff, "%s  %4d %3d %4d  %s",
      (annotate) ? " " : "", frpolys, fp->nmismat, totdifs, abuff);
   dumpnl(); }



/* For building sequence names */

static char *setnames(char *qbf, SIMTYPE simtype, int i1, int i2, int g)
 { switch(simtype) {
     default:  return  "?????";
     case SIM_INNER: case SIM_OUTERPAIR:
        /* A semi-colon separated quoted pair */
        sprintf(qbf, "%s;%s", strtag[i1],strtag[i2]);
        return qbf;
     case SIM_OUTERSEQ:
        return strcpy(qbf,strtag[i1]);
     case SIM_OUTERGROUP:
        /* The column heading says `Group Name' */
        return strcpy(qbf,grouplist[g].gname);
           }}


/* Display global and pairwise fragments */

static void show_blastscores(SIMX *sc)
 { int npairs=sc->npairs, mxlen=sc->bmxlen, valid=FALSE;
   char ftag[4], nbuff[22];  int x,maxsc,scorewid;
   SIMTYPE simtype=sc->simtype;
   PAIRCOMP *pcend, *pc=sc->pairpc;  pcend=pc+npairs;
   set_ftag(ftag, 'G', sc->simtype);
   dumpxnl();
   dumpxtext ("BLAST-like scores (%s) used "
      "in global comparisons (gscale=%d):\n", ftag, gscale);
   maxsc=99;  /* At least two digits */
   for (pc=sc->pairpc; pc<pcend; pc++)
     if ( (x=pc->mxscore)>maxsc)  maxsc=x;
   scorewid=sprintf(nbuff,"%d",maxsc);
   for (pc=sc->pairpc; pc<pcend; pc++)
    { char names[MAXNAMETWO];
      setnames(names, simtype, pc->seq1,pc->seq2,pc->group);
      dumpxtext("  %-*s   ", mxlen,names);
      if (pc->kavalid)
       { char sign='+', *bleep, *bleepl;
         double kanp=log(pc->kmax*npolys);
         if (gscale>0) { bleep="SCORE";  bleepl="Score"; }
         else { bleep="NPOLYS";  bleepl="Npolys"; }
         if (kanp<0.0) { sign='-';  kanp=-kanp; }
         dumptext ("BLAST = %6.4f*%s %c %6.4f    Max%s=%*d\n",
             pc->ll, bleep, sign, kanp, bleepl,
             scorewid,pc->mxscore);
         valid=TRUE; }
      else if (pc->kaerrmess && pc->kainterr==FALSE)
         dumptext("%s\n", pc->kaerrmess);
      else  dumptext("Global scores not valid.\n"); }
   if (valid)
    { dumpxnl();
      if (gscale>0)
       { char *p=(sc->simtype==SIM_INNER) ? "ndifs" : "nmats";
         dumpxtext (
         "where SCORE = (npolys-%s)*1 + %s*(-mismatpen)\n", p,p); }
      else
        dumpxtext ("where NPOLYS is the number "
         "of polymorphisms in the fragment.\n"); }}


static void show_fragdata(SIMX *sc)
 { int npairs=sc->npairs, mxlen=sc->bmxlen, valid=FALSE;
   char ftag[4], nbuff[22];
   int maxsc,maxtotdf,maxpen,maxnfrg;
   size_t scorewid,totdifwid,penaltywid,fragwid;
   SIMTYPE simtype=sc->simtype;
   PAIRCOMP *pcend, *pc=sc->pairpc;  pcend=pc+npairs;
   set_ftag(ftag, 'P', sc->simtype);
   dumpxnl();
   dumpxtext (
   "Maximum pairwise scores (%s), number of discordant sites,\n", ftag);
   if (gscale>0)
      dumpxtext (
      "  mismatch penalty, and the number of stored fragments:\n");
   else
    { dumpxtext ("  and the number of stored fragments:\n");
      dumpxtext ("(Mismatches in fragments not allowed.)\n"); }
   maxsc=maxtotdf=maxpen=maxnfrg=99;  /* At least two digits */
   for (pc=sc->pairpc; pc<pcend; pc++)
    { int x;
      if ( (x=pc->mxscore)>maxsc)  maxsc=x;
      if ( (x=pc->totdifs)>maxtotdf)  maxtotdf=x;
      if (pc->kavalid && (x=-pc->mismatpen)>maxpen)  maxpen=x;
      if ( (x=pc->nfrags)>maxnfrg)  maxnfrg=x; }
   scorewid=sprintf(nbuff,"%d",maxsc);
   totdifwid=sprintf(nbuff,"%d",maxtotdf);
   penaltywid=sprintf(nbuff,"%d",maxpen);
   fragwid=sprintf(nbuff,"%d",maxnfrg);
   for (pc=sc->pairpc; pc<pcend; pc++)
    { char names[MAXNAMETWO];
      setnames(names, simtype, pc->seq1,pc->seq2,pc->group);
      dumpxtext("  %-*s   ", mxlen,names);
      if (pc->kavalid)
       { char nbuff[99];  nbuff[0]=0;
         if (gscale>0)  sprintf(nbuff, "mismatpen=%*d   ",
            penaltywid,-pc->mismatpen);
         dumptext ("Max=%*d   disc=%*d   %s%*d frag%s stored\n",
            scorewid,pc->mxscore, totdifwid,pc->totdifs, nbuff,
            fragwid,pc->nfrags, (pc->nfrags==1) ? " " : "s");
         valid=TRUE; }
      else if (pc->kaerrmess && pc->kainterr==FALSE)
         dumptext("%s\n", pc->kaerrmess);
      else  dumptext("Global scores not valid.\n"); }
#ifdef VTEXT
   dumpxtext("Global fragments stored: %d   "
     "Max fraglist: %d   MaxScore: %.2f\n",
     sc->ngfrags, sc->maxnglob, sc->obsgs);
#endif
   /* If any pair has valid Karlin-Altschul scores */
   if (valid)
    { dumpxnl();
      if (gscale>0)
       { char *p=(sc->simtype==SIM_INNER) ? "ndifs" : "nmats";
         dumpxtext (
         "where SCORE = (npolys-%s)*1 + %s*(-mismatpen)\n", p,p); }
      else  dumpxtext ("where NPOLYS is the number "
              "of polymorphisms in the fragment.\n"); }}


static int totpairfrags (SIMX *sc)
 { int num=0;  PAIRCOMP *pc,*pcend;
   if (sc!=NULL && (pc=sc->pairpc)!=NULL)
    { pcend=pc+(sc->npairs);
      while(pc<pcend)  { num+=pc->nfrags;  pc++; }}
   return num; }


static char poly_alldesc[] =
  "Num Poly  is the number of polymorphic sites in the fragment.\n";

static char *inner_desc[3] =
 { "Num Dif   is the number of mismatches within the fragment.\n",
   "Tot Difs  is the total number of mismatches between two sequences.\n",
   "MisM Pen  is the penalty per mismatch for these two sequences.\n" };

static char *outerseq_desc[3] =
 { "Num Mat   is the number of nonunique sites within the fragment.\n",
   "Tot Mats  is the total number of nonunique sites for that sequence.\n",
   "MisM Pen  is the penalty per nonunique site for this sequence.\n" };

static char *outerpair_desc[3] =
 { "Num Mat   is the number of matches within the fragment.\n",
   "Tot Mats  is the total number of matches between two sequences.\n",
   "MisM Pen  is the penalty per match for these two sequences.\n" };

static char *outergroup_desc[3] =
 {
 "Num Mat   is the number of group-monomorphic sites within the fragment.\n",
 "Tot Mats  is the total number of monomorphic sites for that group.\n",
 "MisM Pen  is the penalty per monomorphic site for that group.\n" };



static int check_outergrp(SIMTYPE simtype)
 { if (simtype==SIM_OUTERGROUP)
    { if (grouptype==G_BLOCS && blocsize==2)  goto skip_outergrp;
      if (grouptype==G_GROUPS)
       { int g;
         for (g=0; g<ngroups; g++)
           if (grouplist[g].dostat && grouplist[g].ns!=2)  break;
         if (g>=ngroups)  goto skip_outergrp; }}
   return FALSE;
skip_outergrp:
   dumpxtext ("All group sizes are 2, so that there is no difference\n");
   dumpxtext ("  between outerseq and outergroup fragments.\n");
   dumpxnl();
   return TRUE; }




static void frag_heading (int global, SIMX *sc, int seqsort)
 { char nbuff[44], n1buff[44], n2buff[44], *pp, **bleep;
   int i, some_restr=FALSE, cropped,lnum, dosims=sc->dosims;
   int usekapw=sc->usekapw, usekaglob=sc->usekaglob;
   int vnpairs=sc->npairs, needsaybc=TRUE;
   int allbc = (bcsims && sc->dosims);
   int havsimpval;  double maxsimpval;
   SIMTYPE simtype=sc->simtype;
   #ifdef VTEXT
   dumpxtext("%s %s fragments (%s, %s):\n",
     (global) ? "Global" : "Pairwise", fragname(simtype),
     numsayc(n1buff,npolys, "polymorphism",FALSE),
     numsayc(n2buff,nbases, "aligned base",FALSE));
   if (global)
    { cropped=sc->gcropped;  lnum=listglobal;  pp="";
      havsimpval=(havsimglobpval && dosims);  maxsimpval=maxsimglobpval; }
   else
    { /* Only called if listpairs==TRUE */
      cropped=sc->pcropped;  lnum=listperpair;  pp=" per pair";
      havsimpval=(havsimpairpval && dosims);  maxsimpval=maxsimpairpval; }
   #else /* If VSHOW */
   dumpxtext("%s fragments (%s, %s aligned):\n", Fragname(simtype),
     numsayc(n1buff,npolys, "polymorphism",FALSE),
     numsayc(n2buff,nbases, "base",FALSE));
   cropped=sc->pcropped;  lnum=listperpair;  pp=" per pair";
   havsimpval=FALSE;  maxsimpval=-1.0;
   #endif
   fragtype_linedescr(simtype,dumpxtext);
   if (cropped && lnum==0)  return;
   if (global)
    { char *what="BC KA";
      if (dosims)
       { if (allbc) what="BC Sim and BC KA";
         else what="Sim and BC KA"; }
      dumpxtext("(%s P-values are corrected "
         "for multiple comparisons.)\n", what); }
   else  dumpxtext("(P-values are not corrected "
           "for multiple comparisons.)\n");
   if (havsimpval)
    { dumpxtext("(Fragments are listed only if Sim P-value <= %g.)\n",
        maxsimpval);  some_restr=TRUE; }
   if (usekapw || usekaglob)
    { if (usekapw && usekaglob)
        dumpxtext("(Fragments are listed only if %s KAPval <= %g)\n",
          (global) ? "Bonferroni-corrected" : "pairwise",
          (global) ? maxkaglobpval : maxkapairpval);
      else
       { int saybc=TRUE;
         if (usekapw)
          { double pval = maxkapairpval*vnpairs;
            if (pval>=1.00)
             { dumpxtext("(Fragments are listed only if pairwise "
                  "KAPval <= %g)\n", maxkapairpval);  saybc=global; }
            else
             { dumpxtext("(Fragments are listed only if pairwise "
                  "KAPval <= %g\n", maxkapairpval);
               dumpxtext("  or equivalently if Bonferroni-corrected "
                  "KAPval <= %g)\n", pval); }}
         else if (usekaglob)
          { double pval = maxkaglobpval/vnpairs;
            dumpxtext("(Fragments are listed only if Bonferroni-corrected "
                "KAPval <= %g\n", maxkaglobpval);
            dumpxtext("  or equivalently if pairwise "
                "KAPval <= %g)\n", pval); }
         if (saybc)
          { dumpxtext("Here BC KA Pvalues are KA Pvalues "
               "multiplied by %d.\n",  vnpairs);
            needsaybc=FALSE; }}
      some_restr=TRUE; }
   if (cropped)
    { dumpxtext("(At most %s listed%s.)\n",
        numsay(nbuff,lnum, "fragment",FALSE), pp);  some_restr=TRUE; }
   dumpxtext(
   "See earlier in the file for %srestrictions on fragments.\n",
      (some_restr) ? "the full " : "");

   #ifdef VTEXT
   if (global && seqsort && sc->ngfrags>0)
     dumpxtext("Fragments sorted "
        "by %s rather than by P-value.\n",
         (simtype==SIM_OUTERGROUP) ? "group" : "sequence");
   if ( (global && sc->ngfrags>0)
         || (!global && totpairfrags(sc)>0) )
     { if (global && showblast)  show_blastscores(sc);
   #else /* sc->nfrags doesn't exist in VSHOW */
   if (totpairfrags(sc)>0)
     { if (showblast)  show_blastscores(sc);
   #endif
       if (showfragdata)  show_fragdata(sc);
       dumpxnl();
       if (expformat)
         dumpxtext("NOTE: 3.14e-04 means 3.14 x 10^{-4} or 0.000314.\n");
       if (dokaglobpvals)
        { char *nam="KA", *type="KA (BLAST-like)";
          if (allbc) { nam="xx"; type="pairwise"; }
          dumpxtext("BC %s means Bonferroni-corrected %s Pvalues.\n",
             nam, type); }

       /* Include a Bonferroni warning */
       dumpxtext("KA Pvalues are NOT corrected "
          "for multiple pairwise comparisons.\n");
       if (needsaybc)
        { char *what=(allbc)?"PW":"KA";
          if (dokaglobpvals)
            dumpxtext("In the following table, BC Pvalues are %s Pvalues "
              "multiplied by %d.\n", what, vnpairs);
          else dumpxtext("Multiply by %d for Bonferroni-corrected "
              "%s Pvalues.\n", vnpairs, what); }

       dumpxtext(poly_alldesc);
       switch (simtype) {
         default:  bleep=inner_desc;  break;
         case SIM_OUTERSEQ: bleep=outerseq_desc; break;
         case SIM_OUTERPAIR: bleep=outerpair_desc; break;
         case SIM_OUTERGROUP: bleep=outergroup_desc; break; }
       for (i=0; i<3; i++)  dumpxtext(bleep[i]);
       if (truncate_pvals)
         dumpxtext("Pvalues are truncated above at 1.00.\n");
       if (AddOffset>0)
         dumpxtext("Alignment offsets are increased by %d.\n",AddOffset);
                 }

  dumpxnl(); }


static int nlisted;

static void fragline (char *ftag, SIMTYPE simtype, int mxlen,
   int seq1, int seq2, int group, int totdifs, int mismatpen,
   double kapval, int npairs, VMFRAG *fp, int vflags)
 { if (no_onlylist || match_onlylist(seq1,seq2))
    { char names[MAXNAMETWO];  nlisted++;
      setnames(names, simtype, seq1,seq2,group);
      if (seq1<0)  seq1=grouplist[group].first;
      if (taboutflag) tab_listline(ftag,names, seq1,seq2,
         kapval,npairs, fp, totdifs,mismatpen, vflags);
      else norm_listline(ftag,names,mxlen, seq1,seq2,
         kapval,npairs, fp, totdifs,mismatpen, vflags); }}


static void end_fraglist(SIMTYPE simtype, int novlap)
 { char nbuff[66], abuff[66];
   sprintf(abuff, "%s fragment", fragname(simtype));
   dumpxtext ("%s listed.\n",  numsay(nbuff, nlisted, abuff, TRUE));
   if (novlap>=0 && gscale>0 && maxoverlap>0 && nlisted>0)
     dumpxtext ("%s discarded.\n",
        numsayc(nbuff,novlap,"overlapping fragment",TRUE)); }



static double get_kpval (VMFRAG *fp, PAIRCOMP *pc)
  { if (pc->kavalid)
 #ifdef VTEXT
      return get_kapval(getkascore(fp->score,pc->ll,pc->kmax));
 #else
      return pow(10, -fp->logpvka);
 #endif
    else return 1.0; }


#ifdef VTEXT

static void list_glist(SIMX *sc, int seqsort);
static void list_bonf_extra(SIMX *sc);

static void list_gfrags(SIMX *sc)
 { if (sc->inmodel)
    { dumpxnl();  dumpxnl();
      if (sc->simtype==SIM_OUTERGROUP
         && check_outergrp(sc->simtype))  return;
      if (sortgfragsboth)
       { list_glist(sc,FALSE);
         dumpxnl();  dumpxnl();
         list_glist(sc,TRUE); }
      else
        list_glist(sc,sortgfragsbyseq); }}


static void list_glist(SIMX *sc, int seqsort)
 { if (sc->inmodel)
    { int gnum, vflags, npairs=sc->npairs, mxlen=sc->gmxlen+1;
      GFRAG *ggfrags=sc->gflist;
      SIMTYPE simtype=sc->simtype;
      frag_heading(TRUE,sc,seqsort);
      if (mxlen<5) mxlen=5;  nlisted=0;
      vflags = HD_GLOBAL;
      if (sc->dosims)  vflags |= HD_SIMS;
      if (bcsims) vflags |= HD_BCSIM;
      if ( (gnum=sc->ngfrags)>0)
       { char ftag[7];  int i, ogroup=-1;
         int (*whichsort)(const void *p, const void *q);
         set_ftag(ftag, 'G', simtype);
         if (taboutflag)  tab_listhead(simtype,vflags);
         else  norm_listhead(mxlen,simtype,vflags);
         /* How should the global list be sorted? */
         if (seqsort) whichsort=sortgfbyseqs;
         else if (bcsims) whichsort=sortgsfrag;
         else  whichsort=sortgkfrag;
         qsort(ggfrags,gnum, sizeof(GFRAG),whichsort);
         for (i=0; i<gnum; i++)
          { GFRAG gg=ggfrags[i];  double kapval;
            kapval=get_kapval(gg.kascore);
            if (seqsort && ngroups>0 &&
              grouptype==G_GROUPS && simtype!=SIM_OUTERGROUP
              && gg.group!=ogroup)
               { dumptext ("Group %s:\n", grouplist[gg.group].gname);
                 ogroup=gg.group; }
            fragline(ftag,simtype,mxlen,
              gg.seq1,gg.seq2,gg.group, gg.totdifs,gg.mismatpen,
              kapval,npairs, &gg.m, vflags); }}
      dumpxnl();
      end_fraglist(simtype,sc->pzovlap);
      sc->nglisted=nlisted;
      if (!bcsims && sc->dosims) list_bonf_extra(sc);  }}



static void lbonf_head(SIMX *sc, int npairs, int mxlen, int vflags)
 { SIMTYPE simtype=sc->simtype;
   dumpxnl();
   dumpxtext(
   "ADDITIONAL PAIRWISE FRAGMENTS with BC Pairwise SimPval < 0.05\n");
   dumpxtext(
   "  OR LISTED GLOBAL FRAGMENTS with significantly better BC SimPval\n");
   dumpxtext("Here BC SimPval is the Pairwise "
      "Sim Pvalue multiplied by %d.\n", npairs);
   if (!listpairs)
    { PAIRCOMP *pairpc=sc->pairpc;
      int i, npairs=sc->npairs, uptoflag=FALSE;
      for (i=0; i<npairs; i++)
        if (pairpc[i].nfrags<3) { uptoflag=TRUE;  break; }
      dumpxtext("(%s pairwise fragments considered per pair.)\n",
        (uptoflag) ? "Up to three" : "Three"); }
   dumpxnl();
   if (taboutflag)  tab_listhead(simtype,vflags);
   else  norm_listhead(mxlen,simtype,vflags); }


static GFRAG *find_gf (GFRAG *gg, int numg, PAIRCOMP *pc,
     int start, int finish)
 { int i, i1=pc->seq1, i2=pc->seq2, g=pc->group;
   for (i=0; i<numg; i++)
     if (gg[i].seq1==i1 && gg[i].seq2==i2 && gg[i].group==g
        && gg[i].m.start==start && gg[i].m.finish==finish)  return gg+i;
   return NULL; }


static void list_bonf_extra(SIMX *sc)
 { char ftag[7];  int gnum=sc->ngfrags, npairs=sc->npairs;
   int np, nfrags, mxlen=sc->gmxlen+1, frowflag=TRUE;
   int siglev=(int)(0.05*numsims), hsiglev=(int)(0.01*numsims);
   GFRAG *gg, *glist=sc->gflist;
   PAIRCOMP *pc;  SIMTYPE simtype=sc->simtype;
   int vflags = HD_SIMS | HD_GLOBAL | HD_BCSIM;
   if (npairs>=siglev)  return;   /* Nothing to do */
   set_ftag(ftag, 'A', simtype);  nlisted=0;
   /* Find interesting pairwise fragments in sc->pairpc: */
   /* Note that interesting fragments can only be found in PAIRCOMPs: */
   /* If you don't CALCULATE pairwise P-values, then they can't be */
   /*   Bonferroni corrected !! */
   /* First, sum over PAIRCOMPs: */
   for (pc=sc->pairpc, np=0; np<npairs; np++, pc++)
     if ( (nfrags=pc->nfrags)>0
       && (no_onlylist || match_onlylist(pc->seq1,pc->seq2)))
      { int k;  MFRAG *fp;
        /* Now sum over fragments: */
        for (fp=pc->mf, k=0; k<nfrags; k++, fp++)
          if (npairs*(*fp->grc)<siglev)
           { int fmatch=FALSE;
             /* Is the fragment in the global list sc->gflist ? */
             if ( (gg=find_gf(glist,gnum, pc,fp->start,fp->finish))!=NULL)
              { if (*gg->m.grc>=siglev ||
                  (npairs*(*fp->grc)<hsiglev && *(gg->m.grc)>=hsiglev))
                    fmatch=TRUE; }
             if (fmatch || gg==NULL)
              { int ngr;  double kapval = get_kpval(fp,pc);
                MFRAG ff=*fp;  ngr=*ff.grc;
                if (ngr==0) ngr=1;  ngr*=npairs;  ff.grc=&ngr;
                if (frowflag)
                 { lbonf_head(sc,npairs,mxlen,vflags);  frowflag=FALSE; }
                fragline(ftag,simtype,mxlen,
                  pc->seq1,pc->seq2,pc->group, pc->totdifs,pc->mismatpen,
                  kapval,npairs, &ff, vflags); }}}
   if (nlisted>0 && frowflag==FALSE)
    { dumpxnl();
      end_fraglist(sc->simtype,-1); }}


#endif   /* VTEXT */


static int firstrowflag;

static void list_pfrags(SIMX *sc)
 { if (sc->inmodel)
    { SIMTYPE simtype=sc->simtype;
      int pgroup=-1, mxlen=sc->pzmxlen+1;
      int npairs=sc->npairs;
      int vflags = (sc->dosims) ? HD_SIMS : 0;
      PAIRCOMP *pc,*pcend;
      dumpxnl();  dumpxnl();
      if (sc->simtype==SIM_OUTERGROUP
         && check_outergrp(sc->simtype))  return;
      frag_heading(FALSE,sc,FALSE);
      if (mxlen<5)  mxlen=5;  nlisted=0;
      pc=sc->pairpc;  pcend=pc+npairs;  firstrowflag=TRUE;
      while (pc<pcend)
        list_pairlist(simtype, vflags, pc++, &pgroup, npairs, mxlen);
      if (firstrowflag==FALSE)  dumpxnl();
      end_fraglist(simtype,sc->pzovlap);
      sc->nplisted=nlisted; }}



static void list_pairlist (SIMTYPE simtype, int vflags,
     PAIRCOMP *pc, int *pgroup, int npairs, int mxlen)
 { int num;
   if ( (num=pc->nfrags)>0)
    { char ftag[7];
      int i, i1=pc->seq1, i2=pc->seq2, group=pc->group;
      int totdifs=pc->totdifs, mismatpen=pc->mismatpen;
      VMFRAG *fp;
      if (firstrowflag)
       { if (taboutflag)  tab_listhead(simtype,vflags);
         else  norm_listhead(mxlen,simtype,vflags);
         firstrowflag=FALSE; }
      set_ftag(ftag, 'P', simtype);
      if (grouptype==G_GROUPS && simtype!=SIM_OUTERGROUP
            && group!=*pgroup)
       { dumptext ("Group %s:\n", grouplist[pc->group].gname);
         *pgroup=pc->group; }
      fp=pc->mf;
      /* Sum over the fragments for this sequence (pair) */
      for (i=0; i<num; i++, fp++)
       { double kapval = get_kpval(fp,pc);
         fragline(ftag,simtype,mxlen,
           i1,i2,group, totdifs,mismatpen,
           kapval,npairs, fp, vflags); }}}



static char *saymax (char *bleep, int havpv, double maxpv)
 { if (havpv)  return getstr("%s=%g", bleep, maxpv);
   else  return getstr("no %s limits", bleep); }


extern char *nx_datatype;

void show_features(void)
 { char *opts[200];  int n=0;
   char nbuff[500], abuff[88], a2buff[88];
   char *saystr=(grouptype==G_GROUPS || nseqs<file_nseqs) ? "Using " : "";
   char *seqstr=(seqtype==PROTEIN) ? "protein sequence" : "DNA sequence";
   opts[n++]=getstr("%s%s%s", saystr,
     numsay(abuff,nseqs, seqstr, (saystr[0]==0)),
       (circular) ? " (circular)" : "");
   if (seqtype==SILENT)
    { opts[n++]=strxdup("SILENT SITES only");
      if (mitcodes)
        opts[n++]=strxdup("using mitochondrial genetic codes"); }

   opts[n++]=strxdup(numsayc(nbuff,nbases, "aligned base",FALSE));
   opts[n++]=strxdup(numsayc(nbuff,npolys, "polymorphism",FALSE));
   if (set_degens(abuff))
     opts[n++]=getstr("degenmenu option%s `%s' set (see documentation)",
       (abuff[1]) ? "s" : "", abuff);
   if (base_range.s)
     opts[n++]=getstr("OFFSETS %d-%d ONLY", base_range.n, base_range.nx);
   if (base_skip_range.s)
     opts[n++]=getstr("SKIP OFFSETS %d-%d",
        base_skip_range.n, base_skip_range.nx);

   opts[n++]=getstr("%s format", get_ffname());
   if (nx_datatype) opts[n++]=getstr("NEXUS datatype=%s", nx_datatype);
   if (gscale>0)
     opts[n++]=getstr("gscale=%d for mismatches",gscale);
   else opts[n++]=strxdup("mismatches in fragments not allowed");

   switch (grouptype) {
     default:  opts[n++]=getstr("no groups defined");  break;
     case G_BLOCS:  opts[n++]=getstr("blocsize=%d", blocsize);
          opts[n++]=strxdup(numsay(nbuff,blocnum, "group", FALSE));
          if (homolflag)
              opts[n++]=strxdup("homologous pairs (between blocks)");
          break;
     case G_GROUPS:
          opts[n++]=getstr("%s defined",
              numsay(nbuff,ngroups, "group", FALSE));
          break; }

   #ifdef VTEXT
   if (numsims>0)
    { opts[n++]=saymax("maxSimPairPval", havsimpairpval,maxsimpairpval);
      opts[n++]=saymax("maxSimGlobPval", havsimglobpval,maxsimglobpval); }
   #endif
   opts[n++]=saymax("maxKAPairPval", havkapairpval, maxkapairpval);
   opts[n++]=saymax("maxKAGlobPval", havkaglobpval, maxkaglobpval);
   if (numsims>0)
     opts[n++]=getstr("simPvals based on %s",
         numsayc(nbuff,numsims,"permutation",FALSE));
   else  opts[n++]=strxdup("no permutations done");

   opts[n++]=strxdup( (bcsims)
     ? "global P-values are Bonferroni-corrected pairwise P-values"
     : "global P-values based on BLAST-like global scores");

   opts[n++]=getstr("fragment limits: minlength=%d", minlength);
   opts[n++]=getstr("minnpolys=%d", minnpoly);
   opts[n++]=getstr("minscore=%d", minscore);
   if (newlglobal) opts[n++]=getstr("global lists=%d", listglobal);
   if (listpairs && listperpair>0)
     opts[n++]=getstr("pairwise max lists=%d", listperpair);
   else  opts[n++]=strxdup("no pairwise fragment lists");
   if (rshuffled)  opts[n++]=strxdup("RANDOMIZED order of sites");
   if (gscale>0)
    { if (maxoverlap>0)
        opts[n++]=getstr("at most %s",
          numsay(abuff,maxoverlap, "overlapping fragment", FALSE));
      else  opts[n++]=strxdup("no limit on overlapping fragments"); }

   /* Only if random numbers have been referred to or used */
   if (newstartseed || numsims>0 || ranflag || dispseed()!=wstartseed)
     opts[n++]=getstr("starting seed=%lu", wstartseed);

   opts[n++]=getstr("%ssites with indels are %s",
     (treat_dels!=TD_SKIP) ? "contiguous ": "", say_indels(treat_dels));
   if (treat_dels==TD_BLOCS)
     opts[n++]=getstr("%s%s",
        numsay(abuff,numdelblocs,"indel block",FALSE),
        (numdelblocs>0) ? (sprintf(a2buff," with up to %d sites each",
                 maxdelblocs),a2buff) : "");
   if (include_monosites)
     opts[n++]=getstr("USING MONOMORPHIC AS WELL AS POLYMORPHIC %s",
       (seqtype==SILENT) ? "CODON POSITIONS" : "SITES");

   if (sortnames)
     opts[n++]=strxdup("sequence names sorted alphabetically");
   if (sortgfragsboth)
     opts[n++]=strxdup(
       "global fragment lists sorted by P-value AND alphabetically");
   else if (sortgfragsbyseq)
     opts[n++]=strxdup(
       "global fragment lists sorted alphabetically");
   if (AddOffset)
     opts[n++]=getstr("all displayed offsets are increased by %d",
       AddOffset);

   if (matchchar)
     opts[n++]=getstr(
      "match (replacement) char in source file is %c (%2XH)",
        matchchar,matchchar);
   else opts[n++]=strxdup("no match (replacement) char in source file");
   if (use_commchars())
    { if (commchars!=NULL && commchars[0])
        opts[n++]=getstr("source lines beginning with one "
           "of `%s' are comments", commchars);
      else opts[n++]=strxdup("no initial-source-line comment flags"); }
   if (newnumbersep)
     opts[n++]=getstr("Number separator as in 123%c456", numbersep);

   /* Now write to textprintf() (which may be dfile) */
   ftoklistf("Options and features: ", opts,n, 1,",",0,
       TOK_FREE_ALL | TOK_ENDPERIOD); }



#ifdef VTEXT

static void set_globparms (SIMX *sc)
 { if (sc->inmodel)
    { int x, gmxlen=0, ghavcirc=FALSE;  SIMTYPE simtype=sc->simtype;
      GFRAG *ggend, *gg=sc->gflist;  ggend=gg+(sc->ngfrags);
      for ( ; gg<ggend; gg++)
       { if ( (x=colwid(simtype,gg->seq1,gg->seq2,gg->group)) > gmxlen)
           gmxlen=x;
         if (gg->m.start<0 || gg->m.finish>npolys)
          { tothavcirc=ghavcirc=TRUE; }}
      sc->gmxlen=gmxlen;  sc->ghavcirc=ghavcirc; }}

void set_totglobparms(void)
 { set_globparms(&cfinx);
   set_globparms(&cfoutpairx);
   set_globparms(&cfoutseqx);
   set_globparms(&cfoutgrpx); }

#endif


static int colwid (SIMTYPE simtype, int i1, int i2, int g)
 { switch (simtype) {
      default: return 0;
      case SIM_INNER:  case SIM_OUTERPAIR:
        return strlen(strtag[i1])+strlen(strtag[i2])+1;
      case SIM_OUTERSEQ:
        return strlen(strtag[i1]);
      case SIM_OUTERGROUP:
        return strlen(grouplist[g].gname); }}


void set_pairparms (SIMX *sc)
 { if (sc->inmodel)
    { int npfrags,bmxlen,mxlen,nlap, pzhavcirc;
      int npairs=sc->npairs;
      PAIRCOMP *pc,*pcend;  SIMTYPE simtype=sc->simtype;
      pc=sc->pairpc;  pcend=pc+npairs;
      npfrags=nlap=bmxlen=mxlen=0;  pzhavcirc=FALSE;
      for ( ;  pc<pcend;  pc++)
       { int i,x,num;
         if ( (x=colwid(simtype,pc->seq1,pc->seq2,pc->group)) > bmxlen)
             bmxlen=x;
         nlap+=pc->novlap;
         if ( (num=pc->nfrags)>0)
          { VMFRAG *pf=pc->mf;  npfrags+=num;  if (x>mxlen) mxlen=x;
            for (i=0; i<num; i++)
              if (pf[i].start<0 || pf[i].finish>npolys)
               { tothavcirc=pzhavcirc=TRUE;  break; }}}
      sc->npairfrags=npfrags;  sc->pzovlap=nlap;
      sc->bmxlen=bmxlen;  sc->pzmxlen=mxlen;
      sc->pzhavcirc=pzhavcirc; }}


void set_totpairparms(void)
 { tothavcirc=FALSE;
   set_pairparms(&cfinx);
   set_pairparms(&cfoutpairx);
   set_pairparms(&cfoutseqx);
   set_pairparms(&cfoutgrpx); }


static void set2_tnlisted (SIMX *sc)
 {
   #ifdef VTEXT
   sc->nglisted=sc->ngfrags;
   #endif
   sc->nplisted=sc->npairfrags; }


static void set_tnlisted(void)
 { set2_tnlisted(&cfinx);
   set2_tnlisted(&cfoutpairx);
   set2_tnlisted(&cfoutseqx);
   set2_tnlisted(&cfoutgrpx); }

static void clear_tnlisted(void)
 { cfinx.nglisted=cfinx.nplisted=0;
   cfoutpairx.nglisted=cfoutpairx.nplisted=0;
   cfoutseqx.nglisted=cfoutseqx.nplisted=0;
   cfoutgrpx.nglisted=cfoutgrpx.nplisted=0; }



static void tb_list (char *tb, int n, char type)
 { if (n==0)  sprintf (tb, "no %c", type);
   else sprintf(tb, "%d %c", n, type); }


static char *gen_namelist (int bni, int ni, int bno, int no,
    int bna, int na, int bng, int ng, int bleepflag)
 { static char fbuff[99];  char tb[8][10];
   int i, len, n=0;  memset(tb,0,sizeof(tb));
   if (bni)  tb_list (tb[n++], ni, 'I');
   if (bno)  tb_list (tb[n++], no, 'O');
   if (bna)  tb_list (tb[n++], na, 'A');
   if (bng)  tb_list (tb[n++], ng, 'G');

   if (n==0)  return "";
   len=sprintf(fbuff, "%s", tb[0]);
   if (n==2)  len+=sprintf(fbuff+len, " and %s", tb[1]);
   else if (n>2) for (i=1; i<n; i++)
     len+=sprintf(fbuff+len, ", %s%s", (i==n-1) ? "and " : "", tb[i]);
   if (bleepflag)  sprintf(fbuff+len, " significant fragments");
   return fbuff; }


#ifdef VTEXT

static char *gnamelist(int allflag, int bleepflag)
 { if (listglobal==0)  return "No fragments listed";
   else
    { if (allflag) set_tnlisted();  /* Set `nlisted'=fraglist lengths */
      return gen_namelist(cfinx.inmodel, cfinx.nglisted,
        cfoutseqx.inmodel, cfoutseqx.nglisted,
        cfoutpairx.inmodel, cfoutpairx.nglisted,
        cfoutgrpx.inmodel, cfoutgrpx.nglisted, bleepflag); }}

#endif

static char *pnamelist(int allflag, int bleepflag)
 { if (listpairs==FALSE || listperpair==0)
     return "No fragments listed";
   else
    { if (allflag) set_tnlisted();  /* Set `nlisted'=fraglist lengths */
      return gen_namelist(cfinx.inmodel, cfinx.nplisted,
        cfoutseqx.inmodel, cfoutseqx.nplisted,
        cfoutpairx.inmodel, cfoutpairx.nplisted,
        cfoutgrpx.inmodel, cfoutgrpx.nplisted, bleepflag); }}



static void fragtype_linedescr (SIMTYPE simtype,
    int (*prfun)(const char *control, ...))
 { char *wpen = (gscale==0) ? "" : " with penalties";
   switch(simtype) {
     default: prfun ("UNKNOWN SIMTYPE = %d !!!\n",simtype);  break;
     case SIM_INNER: prfun (
   "(Inner fragments are runs of matching sites%s.)\n", wpen);  break;
     case SIM_OUTERSEQ:  prfun (
   "(Outer-sequence fragments are runs of %s.)\n",
      (grouptype==G_GROUPS || blocsize>0)
        ? "sites that are unique in that group" :
          ( (gscale==0) ? "unique sites" : "unique sites with penalties"));
            break;
     case SIM_OUTERPAIR: prfun (
   "(Outer-pair fragments are runs of mismatched sites%s.)\n", wpen);  break;
     case SIM_OUTERGROUP:  prfun (
   "(Outer-group frags are based on runs of "
        "within-group polymorphic sites.)\n");  break; }}




static void set_ftag (char *ftag, char tchar, SIMTYPE simtype)
 { char cc='?';
   switch (simtype) {
     case SIM_INNER: cc='I';  break;
     case SIM_OUTERSEQ: cc='O';  break;
     case SIM_OUTERPAIR: cc='A';  break;
     case SIM_OUTERGROUP: cc='G';  break; }
   sprintf(ftag, "%c%c", tchar, cc); }


static void Warn_nofullseqs(const char *dfile, int annotflag)
 { char nbuff[200], sbuff[66];
   char *full="FULL SEQUENCES NOT SAVED!";
   if (r_polyflag)  sprintf(full=nbuff,
     "THE SEQUENCE FILE `%s' IS IN LISTED POLYMORPHISMS FORMAT!",
       shortname(sbuff,seqfile));
   WarnHandler ("%s\n"
     "CANNOT WRITE %s%s!!!\n"
     "RESETTING %s flag.\n", full, shortname(sbuff,dfile),
       (annotflag) ? " WITH NON-ALIGNED OFFSETS" : "",
       (annotflag) ? "'annotate'" : "'dumpjseqs'"); }


/* (***BBB) Package for listing segments */
/*   before and after ends of fragments  */


static void list_pjseqs(SIMX *sc);
static void fragnbhd (int i1, int i2, FRAGOFFS x);
static void init_jbuffs(void);
static void free_jbuffs(void);

/* Write a file with the sequences on either side */
/*   of estimated gene conversion insertion points */
/* Dump_jseqs() is called only from dump_fragments() */

static int dump_jseqs(char *dfile)
 { char nbuff[88];
   if (saveseqs==FALSE || seqbase==NULL || r_polyflag)
    { Warn_nofullseqs(dfile,FALSE);
      return  dumpjseqs=FALSE; }

   /* `Dumpx_open()' opens the output file and initializes */
   /*   file writing functions */
   dumpx_open(dfile,DX_JSEQS);

   /* Write mostly the same heading as in FRAGS output */
   write_dump_heading("Junction sequences",DX_JSEQS);
   frag_summary_heading();
   say_indel_aligned_offsets();

   if (tothavcirc)
    { dumpxtext(
       "`CA' stands for `Circle Around', and means that the fragment\n");
      dumpxtext("  encircles the endpoints.\n"); }

   if (numdelblocs>0)
    { dumpxtext(
        "Note: Mismatches can occur at a run of sites with indels,\n");
      dumpxtext("  here with up to %d sites.\n",
         maxdelblocs); }
   dumpxnl();

   dumpxtext(
    "Up to %s are listed before and after the beginning and end\n",
        numsay(nbuff,jseqlen,"base",FALSE));
   dumpxtext("  of each fragment.\n");
   /* Open some buffers depending on  jseqlen */
   init_jbuffs();

   /* If VTEXT is set, display global fragments */
   clear_tnlisted();
   #ifdef VTEXT
   if (listglobal>0)
    { list_gjseqs(&cfinx);
      list_gjseqs(&cfoutseqx);
      list_gjseqs(&cfoutpairx);
      list_gjseqs(&cfoutgrpx);
      say_written("global", gnamelist(FALSE,FALSE)); }
   else  say_nofrags("global",TRUE);
   #endif

   /* List junction sequences for (pairwise) fragments */
   if (listpairs && listperpair>0)
    { list_pjseqs(&cfinx);
      list_pjseqs(&cfoutseqx);
      list_pjseqs(&cfoutpairx);
      list_pjseqs(&cfoutgrpx);
      #ifdef VTEXT
      say_written("pairwise", pnamelist(FALSE,FALSE));
      #else
      say_written(NULL, pnamelist(FALSE,FALSE));
      #endif
        }
   else  say_nofrags("pairwise",TRUE);

   free_jbuffs();
   dumpx_close(dfile);
   return TRUE; }



/* Functions used in dump_jseqs() */


static void jseq_hline (char *bleep, SIMTYPE simtype)
 { char fbuff[55], cbuff[22];
   if (bleep!=NULL && bleep[0]!=0)
     sprintf(fbuff, "%s %s", bleep, fragname(simtype));
   else  sprintf(fbuff, "%s", Fragname(simtype));
   dumpxtext("%s fragments (%s%s aligned bases):\n",
     fbuff, (simtype==SIM_OUTERGROUP) ? "group names, " : "",
     commll(cbuff,nbases)); }


static int jseq_head (char *bleep, SIMTYPE simtype)
 { int pairflag=(simtype==SIM_INNER || simtype==SIM_OUTERPAIR);
   dumpxnl();
   if (simtype==SIM_OUTERGROUP && check_outergrp(simtype))
     return TRUE;
   jseq_hline(bleep,simtype);
   dumpxtext("(Sites with indels %sare ignored.)\n",
      (pairflag) ? "in both sequences " : "");
   dumpxtext("Display:  Before <Frag...ment> After\n");
   dumpxnl();
   return FALSE; }


static void show_jseq (char *ftag, char *name, int start, int finish,
  int i1, int i2,  double pval, char *pvaltype)
 { char pbuff[22];  int lenf;
   /* typedef struct { int lf,rt,caf; } FRAGOFFS; */
   FRAGOFFS x = offsfrags(start,finish);
   lenf=x.rt-x.lf+1;  if (lenf<=0) lenf+=nbases;
   if (pval>1.00) sprintf(pbuff, ">1.00");
   else  sprintf(pbuff, "=%6.4f", pval);  nlisted++;
   /* Heading line for fragment */
   fprintf(dpff, "%s  %s (%d,%s%d, len=%d, %s Pvalue%s):\n",
     ftag, name, x.lf + AddOffset,
        (x.caf) ? "CA" : "  ", x.rt + AddOffset, lenf, pvaltype, pbuff);
   /* Now the aligned insertion-point neighborhood */
   fragnbhd(i1,i2,x); }


static void end_jseqlist(SIMTYPE simtype)
 { if (nlisted==0)  end_fraglist(simtype,-1);
   else if (nlisted>=3)
    { dumpxnl();
      end_fraglist(simtype,-1); }}


#ifdef VTEXT

static void list_gjseqs(SIMX *sc)
 { if (sc->inmodel)
    { char ftag[5], *pvaltype;
      SIMTYPE simtype=sc->simtype;  GFRAG *ggend,*gg;
      int gnum=sc->ngfrags, npairs=sc->npairs, dosims=sc->dosims;
      if (jseq_head("Global",simtype) || gnum==0)  return;
      set_ftag(ftag,'G',simtype);  nlisted=0;
      pvaltype = (dosims) ? "Sim" : "BC KA";
      /* Loop over global fragments */
      gg=sc->gflist;  ggend=gg+gnum;
      for ( ; gg<ggend; gg++)
       { char q1b[MAXNAMETWO];
         int i1=gg->seq1, i2=gg->seq2, g=gg->group;
         double pval;
         if (no_onlylist || match_onlylist(i1,i2))
          { char *name=fragseqs(q1b,i1,i2,g,simtype);
            if (i1<0)  i1=grouplist[gg->group].first;
            if (dosims)  pval=(double)*gg->m.grc/numsims;
            else  pval=npairs*get_kapval(gg->kascore);
            show_jseq(ftag, name, gg->m.start,gg->m.finish,
                i1,i2, pval, pvaltype); }}
      sc->nglisted=nlisted;
      end_jseqlist(simtype); }}



#endif


static void list_pjseqs(SIMX *sc)
 { if (sc->inmodel)
    { char ftag[5], *pvaltype;
      SIMTYPE simtype=sc->simtype;  PAIRCOMP *pc,*pcend;
      int npairs=sc->npairs, num=sc->npairfrags, dosims=sc->dosims;
    #ifdef VTEXT
      if (jseq_head("Pairwise",simtype) || num==0)  return;
    #else
      if (jseq_head("",simtype) || num==0)  return;
    #endif
      set_ftag(ftag,'P',simtype);  nlisted=0;
      pvaltype = (dosims) ? "Sim" : "KA";
      /* Loop over pairs */
      pc=sc->pairpc;  pcend=pc+npairs;
      for ( ; pc<pcend; pc++)
       { char q1b[MAXNAMETWO];
         int i, i1=pc->seq1, i2=pc->seq2, g=pc->group, n=pc->nfrags;
         VMFRAG *fp=pc->mf;  double pval;
      #ifdef VTEXT
         double lla=pc->ll, kmaxa=pc->kmax;
      #endif
         if (no_onlylist || match_onlylist(i1,i2))
          { char *name=fragseqs(q1b,i1,i2,g,simtype);
            if (i1<0)  i1=grouplist[g].first;
            /* Loop over fragments for that pair */
            for (i=0; i<n; i++, fp++)
             {
            #ifdef VTEXT
               if (dosims)
                 pval=(double)*fp->grc/numsims;
               else
                 pval=get_kapval(getkascore(fp->score,lla,kmaxa));
            #else
               pval=pow(10, (dosims) ? -fp->logpvsim : -fp->logpvka);
            #endif
               show_jseq(ftag, name, fp->start,fp->finish,
                   i1,i2, pval, pvaltype); }}}
      sc->nplisted=nlisted;
      end_jseqlist(simtype); }}




/* Buffers used by fragnbhd() and friends */

static char *bufleft1, *bufleft2;
static char *bufcenter1, *bufcenter2, *bufright1, *bufright2;
static char *bufrev1, *bufrev2;

static void init_jbuffs(void)
 { int n=jseqlen+10;  free_jbuffs();
   bufleft1=mxalloc(n);  bufcenter1=mxalloc(2*n);  bufright1=mxalloc(n);
   bufleft2=mxalloc(n);  bufcenter2=mxalloc(2*n);  bufright2=mxalloc(n);
   bufrev1=mxalloc(n);  bufrev2=mxalloc(n); }


static void free_jbuffs(void)
 { if (bufrev1)
    { free(bufrev2);  free(bufrev1);
      free(bufright2);  free(bufcenter2);  free(bufleft2);
      free(bufright1);  free(bufcenter1);  free(bufleft1);
      bufleft1=bufleft2=bufcenter1=bufcenter2=NULL;
      bufright1=bufright2=bufrev1=bufrev2=NULL; }}


/* Functions used in list_pjseqs() */

static int fragsegl(int k1, int k2, int flen, char *sq1, char *sq2);
static int fragsegr(int k1, int k2, int flen, char *sq1, char *sq2);
static void fragsegc(FRAGOFFS x, int flen, char *sq1, char *sq2);


/* typedef struct { int lf,rt,caf; } FRAGOFFS; */

static void fragnbhd (int i1, int i2, FRAGOFFS x)
 { char *sq1=seqbase[i1], *sq2=(i2>=0) ? seqbase[i2] : NULL;
   int flen=jseqlen;
   x.lf--;  x.rt--;  /* Convert to offsets */
   fragsegc(x, flen,sq1,sq2);
   fragsegl(x.lf,0,flen,sq1,sq2);
   fragsegr(x.rt,nbases,flen,sq1,sq2);
   fprintf(dpff, "  %-*s:  %s%s%s%s%s\n", mxtag,strtag[i1],
     bufleft1, (bufleft1[0]) ? " " : "",
        bufcenter1, (bufright1[0]) ? " " : "", bufright1);
   if (sq2!=NULL)
     fprintf(dpff, "  %-*s:  %s%s%s%s%s\n", mxtag,strtag[i2],
       bufleft2, (bufleft2[0]) ? " " : "",
          bufcenter2, (bufright2[0]) ? " " : "", bufright2); }


static void fragsegc(FRAGOFFS x, int flen, char *sq1, char *sq2)
 { int n,aa,bb;
   /* For the l,r portions of the fragment itself */
   if (x.caf)
    { aa=fragsegr(x.lf-1, nbases, flen,sq1,sq2);
      bb=fragsegl(x.rt+1, 0, flen,sq1,sq2);
      /* Overlap cannot happen since len(frag)<=nbases */
      n=0;  if (aa<nbases)  n+=(nbases-aa);
      if (bb>=0)  n+=(bb+1); }
   else
    { aa=fragsegr(x.lf-1, x.rt, flen,sq1,sq2);
      bb=fragsegl(x.rt+1, x.lf, flen,sq1,sq2);
      n=bb-aa+1; } /* n = rt+1-lf-2*flen = len(frag)-2*flen */
   if (n>=0)  /* A gap or a perfect fit */
    { char dbuff[7];  int i=0;  if (n>4) n=4;
      while(--n>=0)  dbuff[i++]='.';  dbuff[i]=0;
      sprintf(bufcenter1, "<%s%s%s>", bufright1,dbuff,bufleft1);
      if (sq2!=NULL)
      sprintf(bufcenter2, "<%s%s%s>", bufright2,dbuff,bufleft2); }
   else  /* Some overlap from left and right */
    { if (aa<=x.rt)  /* Not a complete overlap */
        fragsegr(x.lf-1, x.rt, 2*flen+10,sq1,sq2);
      sprintf(bufcenter1,"<%s>", bufright1);
      if (sq2!=NULL)
      sprintf(bufcenter2,"<%s>", bufright2); }}


/* Ignoring matched indels, copy up to  flen  characters */
/*   strictly to the LEFT of offset k1 */
/*   while remaining to the right of offset k2 */
/* Return the offset of one base before the last base copied */

static int fragsegl(int k1, int k2, int flen, char *sq1, char *sq2)
 { char *p1,*p2=NULL;
   int i, k0=k1, n=0;
   while (--k0>=0 && k0>=k2 && n<flen)
    { p1=strchr(dump_knownchars,sq1[k0]);
      if (sq2!=NULL)  p2=strchr(dump_knownchars,sq2[k0]);
      if (p1==NULL && p2==NULL)  continue;
      bufrev1[n] = (p1) ? sq1[k0] : '*';
      if (sq2!=NULL)
      bufrev2[n] = (p2) ? sq2[k0] : '*';  n++; }
   /* Now reverse the order */
   for (i=0; i<n; i++)
    { bufleft1[i]=bufrev1[n-i-1];  bufleft2[i]=bufrev2[n-i-1]; }
   bufleft1[n]=bufleft2[n]=0;
   return k0; }

/* Ignoring matched indels, copy up to  flen  characters */
/*   strictly to the RIGHT of offset k1 */
/*   while remaining to the left of offset k2 */
/* Return the offset of one base after the last base copied */

static int fragsegr(int k1, int k2, int flen, char *sq1, char *sq2)
 { char *p1,*p2=NULL;  int k0=k1, n=0;
   while (++k0<nbases && k0<=k2 && n<flen)
    { p1=strchr(dump_knownchars,sq1[k0]);
      if (sq2!=NULL)  p2=strchr(dump_knownchars,sq2[k0]);
      if (p1==NULL && p2==NULL)  continue;
      bufright1[n] = (p1) ? sq1[k0] : '*';
      if (sq2!=NULL)
      bufright2[n] = (p2) ? sq2[k0] : '*';  n++; }
   bufright1[n]=bufright2[n]=0;
   return k0; }




/* Write inner and outer fragments in a custom plot format */

#ifdef VTEXT
static void list_plotglobfrags(SIMX *sc);
#else
extern SIMTYPE simintype, simouttype;
#endif
static void list_plotpairfrags(SIMX *sc);

/* Dump_plot() is called only from dump_fragments() */

static int dump_plot(char *dfile)
 { SIMTYPE osimtype;  char *wtype, *namelist;
   void (*list_plot)(SIMX *sc);
   /* `Dumpx_open()' opens the output file and initializes */
   /*   file writing functions */
   dumpx_open(dfile,DX_PLOT);

   /* Write the heading */
   dumptext("#PLOTLIST\n");
   write_dump_heading("Plot file",DX_PLOT);
   dumpxnl();

   #ifdef VTEXT
   if (pairplot)
    { if (!listpairs || listperpair==0)
       { say_nofrags("pairwise",TRUE);  goto dump_exit; }
      dumpxtext("PAIRWISE FRAGMENTS: %s.\n", pnamelist(TRUE,TRUE));
      list_plot=list_plotpairfrags; }
   else
    { if (listglobal==0)
       { say_nofrags("global",TRUE);  goto dump_exit; }
      dumpxtext("GLOBAL FRAGMENTS: %s.\n", gnamelist(TRUE,TRUE));
      list_plot=list_plotglobfrags; }
   /* List one type of outer fragments */
   osimtype=get_outtype();
   #else
   if (!listpairs || listperpair==0)
    { say_nofrags("pairwise",TRUE);  goto dump_exit; }
   dumpxtext("PAIRWISE FRAGMENTS: %s.\n", pnamelist(TRUE,TRUE));
   list_plot=list_plotpairfrags;
   /* List one type of outer fragments */
   osimtype=simouttype;
   #endif

   dumpxnl();
   dumpxtext ("Fragment offsets and lengths are in %s.\n",
       (seqtype==PROTEIN) ? "codons" : "nucleotides");
   dumpxtext ("Heights are -log10 of Pvalues.\n");
   dumpxtext ("Negative heights are replaced by 0\n");

   /* Now list the fragments */
   clear_tnlisted();
   if (innerpairfrags)
    { dumptext ("\npolylines %s\n", (altcolor) ? "blue" : "green");
      list_plot(&cfinx); }
   if (osimtype!=SIM_NONE)
    { dumptext ("\npolylines %s\n", (altcolor) ? "yellow" : "red");
      list_plot(fragsimx(osimtype)); }

   #ifdef VTEXT
   if (pairplot) { namelist=pnamelist(FALSE,FALSE);  wtype="pairwise"; }
   else { namelist=gnamelist(FALSE,FALSE);  wtype="global"; }
   #else
   namelist=pnamelist(FALSE,FALSE);  wtype="";
   #endif
   say_written(wtype, namelist);

dump_exit:
   dumpx_close(dfile);
   return TRUE; }



/* Functions used in dump_plot() */

static void plot_head (char *bleep, SIMTYPE simtype)
 { dumpxnl();  jseq_hline(bleep,simtype);
   dumpxtext("Display:  No. of points  Nodes(x,y)\n");
   dumpxnl(); }


static void show_plotline (int start, int finish, double pvht)
 { FRAGOFFS x = offsfrags(start,finish);
   x.lf += AddOffset;  x.rt += AddOffset;
   if (pvht<0.0)  pvht=0.0;  nlisted++;
   if (x.caf)
    { fprintf(dpff, "    3      1 %.3f   %4d %.3f   %4d 0\n",
         pvht, x.rt,pvht, x.rt);
      fprintf(dpff, "    3   %4d 0   %4d %.3f   %4d %.3f\n",
         x.lf, x.lf,pvht, nbases,pvht); }
   else
     fprintf(dpff, "    4   %4d 0   %4d %.3f   %4d %.3f   %4d 0\n",
        x.lf, x.lf,pvht, x.rt,pvht, x.rt); }



#ifdef VTEXT

static void list_plotglobfrags(SIMX *sc)
 { if (sc!=NULL && sc->inmodel)
    { SIMTYPE simtype=sc->simtype;  GFRAG *ggend,*gg;
      int gnum=sc->ngfrags, npairs=sc->npairs, dosims=sc->dosims;
      plot_head("Global",simtype);
      if (gnum==0)  return;
      /* Loop over global fragments */
      gg=sc->gflist;  ggend=gg+gnum;  nlisted=0;
      for ( ; gg<ggend; gg++)
       if (no_onlylist || match_onlylist(gg->seq1,gg->seq2))
        { double pvht, pval;
          if (dosims)
           { pval=(double)*gg->m.grc/numsims;
             if (havMinDispPvalue && pval<minDispPvalue)
               pval=minDispPvalue;
             if (pval<1e-80)  pval=1e-80; }
          else  pval=npairs*get_kapval(gg->kascore);
          pvht=-log10(pval);
          show_plotline(gg->m.start, gg->m.finish, pvht); }
      sc->nglisted=nlisted;
      end_jseqlist(simtype); }}

#endif


static void list_plotpairfrags(SIMX *sc)
 { if (sc!=NULL && sc->inmodel)
    { SIMTYPE simtype=sc->simtype;  PAIRCOMP *pc,*pcend;
      int npairs=sc->npairs, num=sc->npairfrags, dosims=sc->dosims;
    #ifdef VTEXT
      plot_head("Pairwise",simtype);
    #else
      plot_head(NULL,simtype);
    #endif
      if (num==0)  return;
      /* Loop over pairs */
      pc=sc->pairpc;  pcend=pc+npairs;  nlisted=0;
      for ( ; pc<pcend; pc++)
       if (no_onlylist || match_onlylist(pc->seq1,pc->seq2))
        { int i, n=pc->nfrags;  VMFRAG *fp=pc->mf;
    #ifdef VTEXT
          double lla=pc->ll, kmaxa=pc->kmax;
    #endif
          /* Loop over fragments for that pair */
          for (i=0; i<n; i++, fp++)
           { double pvht;
          #ifdef VTEXT
             if (dosims)
              { double pval=(double)*fp->grc/numsims;
                if (havMinDispPvalue && pval<minDispPvalue)
                  pval=minDispPvalue;
                if (pval<1e-80)  pval=1e-80;
                pvht=-log10(pval); }
             else  pvht=getlogkapval(fp->score,lla,kmaxa);
          #else
             pvht = (dosims) ? fp->logpvsim : fp->logpvka;
          #endif
             show_plotline(fp->start, fp->finish, pvht); }}
      sc->nplisted=nlisted;
      end_jseqlist(simtype); }}




/* (***CCC) Utilities used in dump files */
/*    (show_seqnames(), show_polysites() etc.) */


static void show_gline(int g, char **toks)
 { char ghead[MAXNAMEBUFF];
   SEQNAME *sq=grouplist[g].sqn;  int ns=grouplist[g].ns;
   int i,n, nsep=(mxtag>5) ? 2 : 1;
   sprintf(ghead, "  %s (%d): ", grouplist[g].gname, ns);
   for (i=0; i<ns; i++)  toks[i]=sq[i].name;  n=ns;
   if (grouplist[g].gen_poly_only)
     toks[n++]="(Special group; generate polys only)";
   else if (grouplist[g].dostat==FALSE)
     toks[n++]="(Statistics not computed)";
   ftoklistf(ghead, toks,n, nsep,NULL,4, TOK_NONE); }


void show_seqnames(void)
 { int ngr=0;
   if (grouptype==G_GROUPS)
    { int g, genpoly=-1;  /* Is there a ``GenPolyOnly'' group? */
      char **toks=cxalloc(nseqs+3,sizeof(char *));
      textprintf ("Groups:\n");
      /* Loop over groups */
      for (g=0; g<ngroups; g++)
       { if (grouplist[g].gen_poly_only) genpoly=g;
         else { show_gline(g,toks);
                if (grouplist[g].dostat) ngr++; }}
      if (genpoly>=0)  show_gline(genpoly,toks);
      free(toks);
      if (seqs_notused.n>0)
       { int nsep=(mxtag>5) ? 2 : 1;
         ftoklistf("Sequences not in group list: ",
            seqs_notused.s,seqs_notused.n, nsep,NULL,0, TOK_NONE); }}
   else
    { ngr=blocnum;  show_all_seqnames("Sequence names:  "); }
   if (grouptype==G_GROUPS || blocsize!=0)
    { char nbuff[44];  numsay(nbuff,ngr, "group",FALSE);
      if (homolflag)  textprintf ("Looking at "
           "homologous fragments within groups (%s).\n", nbuff);
      else  textprintf ("Looking at within-group fragments only (%s)\n",
               nbuff); }}



void show_all_seqnames(char *bleep)
 { int i,n,nsep=1;
   char **str=cxalloc(2*nseqs+5,sizeof(char *));
   for (i=0; i<nseqs; i++)
    { char *p=strtag[i];
      if (strlen(p)>5) { nsep=2;  break; }
      while (*p) if (!isalnum(*p++)) { nsep=2;  break; }}
   for (n=i=0; i<nseqs; i++)
    { if (blocsize>0 && i>0 && (i%blocsize)==0)  str[n++]="";
      str[n++]=strtag[i]; }
   ftoklistf(bleep, str,n, nsep,NULL,0, TOK_NONE);
   free(str); }


/* Display  seq_list, seq_skip_list, filepath if not NULL: */
/* These lists are in ``unquoted'' form. */

static void show_one_list(char *bleep, char *bleep2, STRLIST a)
 { if (a.n>0)
    { int i, nsep=1;  TOKLIST tokl;
      memset(&tokl,0,sizeof(tokl));
      if (bleep2!=NULL) { addtokl(&tokl, bleep2, TOK_NONE);  nsep=2; }
      else  for (i=0; i<a.n; i++)
       { if (strlen(a.s[i])>5) { nsep=2;  break; }}
      for (i=0; i<a.n; i++)  addtokl(&tokl, a.s[i], TOK_QUOTE);
      ftoklistex(bleep, &tokl, nsep,NULL,0, TOK_FREE); }}


static void show_seq_path_lists(void)
 { if (seq_skip_list.n>0)
    { show_one_list("Specifying sequences: ", NULL, seq_list);
      show_one_list( (seq_skip_list.n==1) ?
        "Skipping sequence: " : "Skipping sequences: ",
          NULL, seq_skip_list); }
   show_one_list("Paths for reading files: ",
     "(Default directory)", filepath); }


void show_unalens(PRINTFUNP prfun, PRINTFUNP prfun2)
 { char *kn=set_offtypes();
   prfun(
     "UNALIGNED LENGTHS for sequences in %s (aligned bases: %d):\n",
         seqfile, nbases);
   prfun ("Bases other than `%s' are treated as indels.\n", kn);
   if (seqbase!=NULL)
    { int i,j,n;
      for (i=0; i<nseqs; i++)
       { char c1buff[22], c2buff[22];
         char *s=seqbase[i];  int len=strlen(seqbase[i]);
         for (n=j=0; j<len; j++) if (strchr(kn,s[j]))  n++;
         prfun ("  %-*s   ", mxtag,strtag[i]);
         prfun2 ("Unaligned length: %4s   Indels: %3s\n",
           commll(c1buff,n),
           (len==n) ? " None" : commll(c2buff,len-n)); }}}


/* Display polymorphisms */

static unsigned char gstar (unsigned char c)
 { return ( ('A'<=c && c<='Z') || ('0'<=c && c<='9')
      || ('a'<=c && c<='z') ) ? c : '*'; }


static unsigned char getcons(const unsigned char *site)
 { int i, ic, vc, cnt[256];
   memset(cnt, 0, sizeof(cnt));
   for (i=0; i<nseqs; i++)  cnt[site[i] & 0xff]++;
   vc = cnt[ic=0];
   /* Return the first nonzero site-value with a plurality */
   for (i=1; i<256; i++)  if (vc<cnt[i])  vc=cnt[ic=i];
   return gstar(ic); }


/* Note: A row of sites by itself in a spreadsheet shell */
/*   can apparently invoke a spreadsheet ``#NAME?'' function. */
/*   Thus, the following should NOT be used with prfun,prfun2 */

void show_polysites(void)
 { int k,klen;
   textprintf ("\n");
   textprintf ("The %spolymorphic sites:\n",
       (numsims>0) ? "permuted " : "");
   if (numdelblocs)  textprintf ("(Sites with 1234... are "
         "blocks of sites with unrecognized bases.)\n");
   if (have_silent_multsites)  textprintf ("(Sites with 1234... are "
         "irregular silent sites with mult.poly.sites.)\n");
   if (polysites[0].loca<=0)  textprintf ("(`CA' stands for `circle "
         "around': some blocks encircle the endpoints.)\n");
   if (AddOffset>0)
     textprintf ("Alignment offsets are increased by %d.\n",AddOffset);

   /* Display the polymorphisms in a graphic: */
   if ( (klen=72-mxtag) < 3)  klen=70;
   for (k=0; k<npolys; k+=klen)
    { unsigned char cons[80], *ca="";  int a,i,j,k2,na;
      if ( (k2=k+klen) > npolys)  k2=npolys;
      if ( (na=polysave[k].loca)<=0) { na+=nbases;  ca="CA"; }
      textprintf ("\n");
      textprintf ("#%d (Offset %s%d) to #%d (Offset %d):\n",
        k+1, ca, na + AddOffset, k2, polysave[k2-1].locb + AddOffset);
      if (matchchar!=0)
       { for (a=0, j=k; j<k2; j++, a++)
           cons[a]=getcons(polysave[j].site);  cons[a]=0;
         textprintf ("%-*s  %s\n", mxtag,"!", cons); }
      for (i=0; i<nseqs; i++)
       { unsigned char line[80];
         if (matchchar!=0)
           for (a=0, j=k; j<k2; j++, a++)
            { unsigned char cc=polysave[j].site[i];
              line[a] = (cc==cons[a]) ? matchchar : gstar(cc); }
         else
           for (a=0, j=k; j<k2; j++, a++)
             line[a] = gstar(polysave[j].site[i]);
         line[a]=0;
         textprintf ("%-*s  %s\n", mxtag,strtag[i], line); }}

   /* Now display the offsets: */
   textprintf ("\n");
   for (k=0; k<npolys; k+=klen)
    { int j,k2;  TOKLIST alist;  memset(&alist,0,sizeof(alist));
      if ( (k2=k+klen) > npolys)  k2=npolys;
      if (k>0)  textprintf("\n");
      textprintf ("%s#%d to #%d:\n",
        (k==0) ? "Offsets for polymorphisms: " : "", k+1, k2);
      for (j=k; j<k2; j++)
       { char nbuff[55];
         if (polysave[j].loca==polysave[j].locb)
           sprintf(nbuff, "%d", polysave[j].loca + AddOffset);
         else if (polysave[j].loca<=0)
           sprintf(nbuff, "(%d,CA%d)",
               polysave[j].loca+nbases + AddOffset,
               polysave[j].locb + AddOffset);
         else
           sprintf(nbuff, "(%d,%d)",
               polysave[j].loca + AddOffset,
               polysave[j].locb + AddOffset);
         addtokl(&alist, strxdup(nbuff), TOK_FREE); }
      ftoklistex("  ", &alist, 1,NULL,0, TOK_FREE); }}



static void show_mpv_head(const char *bleep)
 { char nbuff[299];
   int len=sprintf(nbuff,"%s",bleep);
   if (nseqs<file_nseqs || grouptype!=G_SIMPLE)
    { char abuff[44];
      len+=sprintf(nbuff+len, " (");
      if (nseqs<file_nseqs)
        len+=sprintf(nbuff+len, "%s%s", numsay(abuff,nseqs, "sequence",FALSE),
        (grouptype!=G_SIMPLE) ? ", " : "");
      if (grouptype!=G_SIMPLE)
       { int g,ngr;
         for (ngr=g=0; g<ngroups; g++) if (grouplist[g].dostat) ngr++;
         len+=sprintf(nbuff+len, numsay(abuff, ngr, "group", FALSE)); }
      if (grouptype==G_BLOCS)
       { if (len>56) { textprintf("%s\n",nbuff);  len=sprintf(nbuff,"  "); }
         len+=sprintf(nbuff+len, " with blocksize %d", blocsize); }
      len+=sprintf(nbuff+len, ")"); }
   len+=sprintf(nbuff+len, ":");
   if (tabfileflag)
    { if (len>48) { textprintf("%s\n",nbuff);  len=sprintf(nbuff," "); }
      sprintf(nbuff+len, "  Not found in a tab file."); }
   textprintf("%s\n", nbuff); }



static char *show_obs (char *nbuff, double ogs, int obv)
 { if (obv)
     { sprintf(nbuff, "%7.3f", ogs);  return nbuff; }
   else  return "Not found"; }

#ifdef VTEXT

/* Not needed unless there are global permutations */
/* Definitions of SIMX etc. read only if VTEXT is defined */

/* Show maxima and minima of simulations, etc: */

/* Find standard deviations etc. */

static double sdev (double sum1, double sum2, long nn)
 { double x;
   if (nn < 2L
     || (x = sum2 - sum1*sum1/nn) <= 0.0)  return 0.0;
   return  sqrt(x/(nn-1)); }

static double sdevam (double gval, double sum1, double sd, long nn)
 { return  (nn<2L || sd == 0.0)
      ? 0.0 : (gval - sum1/nn) / sd; }


static void mm_stats(SIMX *sc)
 { double sd,sd2;  sd=sd2=0.0;
   if (sc->inmodel)
    { sd=sdev(sc->sum1, sc->sum2, numsims);
      sd2=sdevam(sc->obsgs, sc->sum1, sd, numsims); }
   sc->sdev=sd;  sc->sdevam=sd2; }


void mm_allstats(void)
 { mm_stats(&cfinx);  mm_stats(&cfoutseqx);
   mm_stats(&cfoutpairx);  mm_stats(&cfoutgrpx);  }


static void show_mmline (char *bleep, double xi, double xo,
    double xa, double xg)
 { char lbuff[299];  int len=0, hd=26;
   if (innerpairfrags && cfinx.dosims)
    { len=sprintf(lbuff, "%-*.*s%6.2f", hd,hd,bleep, xi);
      bleep="";  hd=14; }
   if (outerseqfrags && cfoutseqx.dosims)
    { len+=sprintf(lbuff+len, "%-*.*s%6.2f", hd,hd,bleep, xo);
      bleep="";  hd=14; }
   if (outerpairfrags && cfoutpairx.dosims)
    { len+=sprintf(lbuff+len, "%-*.*s%6.2f", hd,hd,bleep, xa);
      bleep="";  hd=14; }
   if (outergrpfrags && cfoutgrpx.dosims)
      sprintf(lbuff+len, "%-*.*s%6.2f", hd,hd,bleep, xg);
   textprintf ("%s\n", lbuff); }


static void show_immline (char *bleep, long xi, long xo,
    long xa, long xg)
 { char lbuff[299];  int len=0, hd=26;
   if (innerpairfrags && cfinx.dosims)
    { len=sprintf(lbuff, "%-*.*s%6ld", hd,hd,bleep, xi);
      bleep="";  hd=14; }
   if (outerseqfrags && cfoutseqx.dosims)
    { len+=sprintf(lbuff+len, "%-*.*s%6ld", hd,hd,bleep, xo);
      bleep="";  hd=14; }
   if (outerpairfrags && cfoutpairx.dosims)
    { len+=sprintf(lbuff+len, "%-*.*s%6ld", hd,hd,bleep, xa);
      bleep="";  hd=14; }
   if (outergrpfrags && cfoutgrpx.dosims)
      sprintf(lbuff+len, "%-*.*s%6ld", hd,hd,bleep, xg);
   textprintf ("%s\n", lbuff); }


void show_maxmeansims(int headflag)
 { char lbuff[299];  int len;
   if (numsims==0
        || (cfinx.obvalid==FALSE && cfoutpairx.obvalid==FALSE &&
            cfoutseqx.obvalid==FALSE && cfoutgrpx.obvalid==FALSE)
        || (cfinx.dosims==FALSE && cfoutpairx.dosims==FALSE &&
            cfoutseqx.dosims==FALSE && cfoutgrpx.dosims==FALSE))  return;
   if (headflag)  show_restricts();
   textprintf ("Simulation results for the maximum global score:\n");
   len=sprintf(lbuff, "Observed:   ");
   if (innerpairfrags && cfinx.dosims)
      len+=sprintf(lbuff+len, "%6sINNER = %6.2f", "", cfinx.obsgs);
   if (outerseqfrags && cfoutseqx.dosims)
      len+=sprintf(lbuff+len, "%5sOUTRSQ = %6.2f", "", cfoutseqx.obsgs);
   if (outerpairfrags && cfoutpairx.dosims)
      len+=sprintf(lbuff+len, "%5sOUTRPR = %6.2f", "", cfoutpairx.obsgs);
   if (outergrpfrags && cfoutgrpx.dosims)
      sprintf(lbuff+len, "%5sOUTRGP = %6.2f", "", cfoutgrpx.obsgs);
   textprintf ("%s\n", lbuff);
   mm_allstats();
   show_mmline ("Max of sims:",
      cfinx.bestgs, cfoutseqx.bestgs, cfoutpairx.bestgs, cfoutgrpx.bestgs);
   show_mmline ("Mean of sims:",
      cfinx.sum1/numsims, cfoutseqx.sum1/numsims,
      cfoutpairx.sum1/numsims, cfoutgrpx.sum1/numsims);
   show_mmline ("S.D. of sims:",
      cfinx.sdev, cfoutseqx.sdev, cfoutpairx.sdev, cfoutgrpx.sdev);
   show_mmline ("Data is S.D. above Mean:",
      cfinx.sdevam, cfoutseqx.sdevam, cfoutpairx.sdevam, cfoutgrpx.sdevam);
   show_immline ("Num.sims. >= observ.:",
      cfinx.nabove, cfoutseqx.nabove, cfoutpairx.nabove, cfoutgrpx.nabove);
          }


static void dump_mmline (char *bleep,
     double xi, double xo, double xa, double xg)
 { dumpxtext("%s",bleep);  dumptext("");  dumptext("");
   if (innerpairfrags && cfinx.dosims) dumpnumform("%6.2f",xi);
   if (outerseqfrags && cfoutseqx.dosims) dumpnumform("%6.2f",xo);
   if (outerpairfrags && cfoutpairx.dosims) dumpnumform("%6.2f",xa);
   if (outergrpfrags && cfoutgrpx.dosims) dumpnumform("%6.2f",xg);
   dumpnl(); }


static void dump_maxmeansims(void)
 { if (numsims==0
        || (cfinx.obvalid==FALSE && cfoutpairx.obvalid==FALSE &&
            cfoutseqx.obvalid==FALSE && cfoutgrpx.obvalid==FALSE)
        || (cfinx.dosims==FALSE && cfoutpairx.dosims==FALSE &&
            cfoutseqx.dosims==FALSE && cfoutgrpx.dosims==FALSE))  return;
   dumpxtext("Simulation results for the maximum global score:\n");
   dumpxtext("");  dumptext("");  dumptext("");
   if (innerpairfrags && cfinx.dosims)      dumptext("INNER");
   if (outerseqfrags && cfoutseqx.dosims)   dumptext("OUTERSEQ");
   if (outerpairfrags && cfoutpairx.dosims) dumptext("OUTERPAIR");
   if (outergrpfrags && cfoutgrpx.dosims)   dumptext("OUTERGRP");  dumpnl();
   mm_allstats();
   dump_mmline("Observed:",
     cfinx.obsgs, cfoutseqx.obsgs, cfoutpairx.obsgs, cfoutgrpx.obsgs);
   dump_mmline("Max of sims:",
     cfinx.bestgs, cfoutseqx.bestgs, cfoutpairx.bestgs, cfoutgrpx.bestgs);
   dump_mmline("Mean of sims:",
     cfinx.sum1/numsims, cfoutseqx.sum1/numsims,
     cfoutpairx.sum1/numsims, cfoutgrpx.sum1/numsims);
   dump_mmline("S.D. of sims:",
     cfinx.sdev, cfoutseqx.sdev, cfoutpairx.sdev, cfoutgrpx.sdev);
   dump_mmline("Data is S.D. above Mean:",
     cfinx.sdevam, cfoutseqx.sdevam, cfoutpairx.sdevam, cfoutgrpx.sdevam);
   dump_mmline("Num.sims. >= observ.",
     cfinx.nabove, cfoutseqx.nabove, cfoutpairx.nabove, cfoutgrpx.nabove); }



static void dumpsimtable (char *bleep, SIMX *vv)
 { if (vv->inmodel)
    { char *head, tbuff[99], nbuff[44];  static int hd;
      if (firstrowflag)
       { head="SCORE";  firstrowflag=FALSE;
         hd=8;  if (strlen(bleep)>6)  hd++;
         textprintf ("%-*.*s %6s %12s   %10s %11s\n",
           hd,hd,bleep, "  Max ", "Sim  ", "S.D.s above", "S.D. of");
         textprintf ("%-*.*s %6s %12s   %10s %11s\n",
           hd,hd," frags", " Score", "P-value", "sim. mean", "sims"); }
      else
       { textprintf ("%s\n", bleep);
         head = " frags"; }
      sprintf (tbuff, "%-*.*s%s",
        hd,hd,head, show_obs(nbuff,vv->obsgs,vv->obvalid));
      if (vv->dosims)
       { double sd = sdev(vv->sum1,vv->sum2,numsims);
         textprintf ("%s %11.4f   %10.2f  %11.4f\n",
           tbuff, (double)vv->nabove/numsims,
             sdevam(vv->obsgs,vv->sum1,sd,numsims), sd); }
      else
        textprintf ("%s   No simulations done\n", tbuff); }}




static void tabsimtable (char *bleep, SIMX *vv)
 { if (vv->inmodel)
    { char *head, nbuff[44];
      if (firstrowflag)
       { char *aline[12], *bline[12];  int i, na=0;
         head = "SCORE";  firstrowflag=FALSE;
         aline[na]=bleep;      bline[na++]="  frags";
         aline[na]="Global";   bline[na++]="Score";
         aline[na]="Sim";      bline[na++]="P-value";
         aline[na]="S.D.s above";  bline[na++]="sim. mean";
         aline[na]="S.D. of";  bline[na++]="sims";
         dumpxtext(aline[0]);
         for (i=1; i<na; i++)  dumptext(aline[i]);  dumpnl();
         dumpxtext(bline[0]);
         for (i=1; i<na; i++)  dumptext(bline[i]);  dumpnl(); }
      else
       { dumpxtext("%s\n",bleep);  head = "  frags"; }
      dumpxtext(head);
      dumptext( show_obs(nbuff,vv->obsgs,vv->obvalid));
      if (vv->dosims)
       { double sd = sdev(vv->sum1,vv->sum2,numsims);
         dumpnumform ("%11.4f", (double)vv->nabove/numsims);
         dumpnumform ("%10.2f", sdevam(vv->obsgs,vv->sum1,sd,numsims));
         dumpnumform ("%11.4f", sd); }
      else  dumptext("No simulations done");
      dumpnl(); }}


#endif  /* VTEXT */


void show_vmainpvals(void)
 { show_mainpvals(DX_NONE); }


static void show_mainpvals (DX_TYPE dx_type)
 { char nbuff[299], abuff[44];
   if (numsims>0 || ranflag)
    { textprintf ("\n");
      textprintf ("The starting random number seed is  %lu.\n", wstartseed); }
   if (numsims>0)
     textprintf ("Simulated P-values are based on %s permutations.\n",
       commll(abuff,numsims));
   else if (dx_type!=DX_NONE)
    { textprintf ("\n");
      textprintf ("No permutations were done.\n"); }
   show_mpv_head("Maximum BLAST-like scores");
   if (tabfileflag)  return;
 #ifdef VTEXT
   if (numsims>0)
    { void (*prtab) (char *, SIMX *) =
         (taboutflag) ? tabsimtable : dumpsimtable;
     /* Display global scores in a vertical format */
      firstrowflag=TRUE;
      prtab("Inner",&cfinx);
      prtab("OuterSeq",&cfoutseqx);
      prtab("OuterPair",&cfoutpairx);
      prtab("OuterGrp",&cfoutgrpx); }
   else
 #endif
     /* Display the scores in a horizontal format */
    { int i,ncols=0;
      if (innerpairfrags) ncols++;  if (outerpairfrags) ncols++;
      if (outerseqfrags) ncols++;  if (outergrpfrags) ncols++;
      if (taboutflag)
       { if (innerpairfrags)  dumptext("Inner");
         if (outerseqfrags)  dumptext("OuterSeq");
         if (outerpairfrags)  dumptext("OuterPair");
         if (outergrpfrags)  dumptext("OuterGrp");
         dumpnl();
         for (i=0; i<ncols; i++)  dumptext(" frags");
         dumpnl();
         if (innerpairfrags)
            dumptext(show_obs(abuff, cfinx.obsgs, cfinx.obvalid));
         if (outerseqfrags)
            dumptext(show_obs(abuff, cfoutseqx.obsgs, cfoutseqx.obvalid));
         if (outerpairfrags)
            dumptext(show_obs(abuff, cfoutpairx.obsgs, cfoutpairx.obvalid));
         if (outergrpfrags)
            dumptext(show_obs(abuff, cfoutgrpx.obsgs, cfoutgrpx.obvalid));
         dumpnl(); }
      else
       { int len=0;
         if (innerpairfrags)  len=sprintf(nbuff, "%-11s", "Inner");
         if (outerseqfrags)  len+=sprintf(nbuff+len, "%-11s", "OuterSeq");
         if (outerpairfrags)  len+=sprintf(nbuff+len, "%-11s", "OuterPair");
         if (outergrpfrags)    sprintf(nbuff+len, "%-11s", "OuterGrp");
         textprintf("%s\n",nbuff);
         for (len=i=0; i<ncols; i++)
            len+=sprintf(nbuff+len, "%-11s", " frags");
         textprintf("%s\n",nbuff);
         len=0;
         if (innerpairfrags) len=sprintf(nbuff, "%-11s",
             show_obs(abuff, cfinx.obsgs, cfinx.obvalid));
         if (outerseqfrags) len+=sprintf(nbuff+len, "%-11s",
             show_obs(abuff, cfoutseqx.obsgs, cfoutseqx.obvalid));
         if (outerpairfrags) len+=sprintf(nbuff+len, "%-11s",
             show_obs(abuff, cfoutpairx.obsgs, cfoutpairx.obvalid));
         if (outergrpfrags)   sprintf(nbuff+len, "%-11s",
             show_obs(abuff, cfoutgrpx.obsgs, cfoutgrpx.obvalid));
         textprintf("%s\n",nbuff); }}}


